# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Contains classes for validation for commonly used
metrics in NeST. For eg., Bandwidth and Delay
"""

import re
import os
from nest.exception import DistributionOptionError
from .input_validator import input_validator


class Metric:
    """
    Base type for metrics. Metrics are of the format <value><unit>.

    For eg., `Bandwidth` type ("10mbit") derives from this class. Here,
    value="10" and unit="mbit"
    """

    valid_units = []

    @input_validator
    def __init__(
        self, metric: str, value_regex=r"[0-9]+(\.[0-9]+)?", unit_regex=r"[A-Za-z]+"
    ):
        """
        Constructor for Metric. Validates if `metric` is in <value><unit>
        format.

        Parameters
        ----------
        metric: str
            Value of the metric
        """
        if not re.fullmatch(value_regex + unit_regex, metric):
            raise ValueError(
                f"{metric} is not a valid metric. Metrics must be of the format "
                f"<value><unit>"
            )

        # Extract value and unit
        self._value = re.search(value_regex, metric).group()
        self._unit = re.search(unit_regex, metric).group()

        self._string_value = metric

    @property
    def value(self):
        """
        Get the value of the metric (without unit)
        """
        return self._value

    @property
    def unit(self):
        """
        Get the unit of the metric
        """
        return self._unit

    @property
    def string_value(self):
        """
        Get string value of the mrtic
        """
        return self._string_value

    @staticmethod
    def allowed_type_cast():
        """
        Indicate str can be typecasted into Metric type
        """
        return [str]

    def __repr__(self):
        """
        String representation of Metric
        """
        classname = self.__class__.__name__
        return f"{classname}({self.string_value!r})"


class Bandwidth(Metric):
    """
    Validates that the bandwidth is in expected format.
    """

    # fmt: off

    valid_units = [
        "bit", "kbit", "kibit", "mbit", "mibit", "gbit", "gibit", "tbit", "tibit",
        "bps", "kbps", "kibps", "mbps", "mibps", "gbps", "gibps", "tbps", "tibps",
    ]

    # fmt: on

    def __init__(self, bandwidth: str):
        """
        Parameters
        ----------
        bandwidth: str
            The value of bandwidth
        """
        super().__init__(bandwidth)

        if self.unit not in Bandwidth.valid_units:
            raise ValueError(f"{self.unit} is not a valid unit for bandwidth.")


class Delay(Metric):
    """
    Validates that the delay is in expected format.
    """

    valid_units = ["s", "sec", "secs", "ms", "msec", "msecs", "us", "usec", "usecs"]

    def __init__(self, delay: str):
        """
        Parameters
        ----------
        delay: str
            The value of delay
        """
        super().__init__(delay)

        if self.unit not in Delay.valid_units:
            raise ValueError(f"{self.unit} is not a valid unit for delay.")

        # Stores the value in milliseconds
        self._value_in_msec = None

        if self.unit in ["us", "usec", "usecs"]:
            self._value_in_msec = str(float(self.value) / 1000)
        elif self.unit in ["ms", "msec", "msecs"]:
            self._value_in_msec = self.value
        else:
            self._value_in_msec = str(float(self.value) * 1000)

    def __add__(self, other):
        """
        Add `Delay` objects and create a new `Delay` object
        with milliseconds unit (by default).
        """
        value_in_msec = float(self._value_in_msec) + float(other._value_in_msec)
        return Delay(f"{value_in_msec}ms")


class Percentage(Metric):
    """
    Validates that the percentage is in expected format.
    """

    def __init__(self, percentage: str):
        """
        Parameters
        ----------
        percentage: str
            The value of percentage
        """
        super().__init__(percentage, unit_regex=r"%")

        if not 0 <= float(self.value) <= 100:
            raise ValueError(f"{self.string_value} is not a valid percentage.")


class Distribution:
    """
    Validate the delay distribution input
    """

    # stores the list of distribution files exist on the path '/usr/lib/tc'
    valid_options = []

    with open("/etc/os-release") as f:
        machine = {}
        for lines in f:
            k, v = lines.rstrip().split("=")
            machine[k] = v.strip('"')

    if machine["NAME"] == "Arch Linux":
        files = [f for f in os.listdir("/usr/share/tc") if f.endswith(".dist")]
    else:
        files = [f for f in os.listdir("/usr/lib/tc") if f.endswith(".dist")]

    for i in files:
        valid_options.append(os.path.splitext(i)[0])

    @input_validator
    def __init__(
        self,
        distribution: str,
    ):
        """
        Parameters
        -----------
        distribution: str
            The delay distribution options: uniform | normal | pareto | paretonormal | experimental
        """
        self._option = distribution

        if self._option not in Distribution.valid_options:
            raise DistributionOptionError(
                "Please set a valid delay distribution option."
                f"\nValid delay distribution options are:{Distribution.valid_options}"
            )

    @property
    def option(self):
        """
        Get the delay distribution option
        """
        return self._option

    @staticmethod
    def allowed_type_cast():
        """
        Indicate str can be typecasted into Metric type
        """
        return [str]



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""CoAP commands"""

import os

from .exec import exec_exp_commands, exec_subprocess_in_background

# pylint: disable=too-many-arguments
def run_coap_client(ns_name, destination_ip, ipv6, coap_options, out, err):
    """
    Run CoAP client program

    Parameters
    ----------
    ns_name : str
        Name of the client namespace
    destination_ip : str
        IP address of the destination namespace
    ipv6 : bool
        Indicates whether the IP address is IPv6 or not
    coap_options : str
        Options string for running the CoAP client application
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    # Here, the path to the current file being run `coap.py` is taken
    # and used to run `coap_client.py` in the same directory. This is
    # done using the `__file__` dunder variable and the `os` library.
    path_to_curr_module = os.path.dirname(os.path.abspath(__file__))
    path_to_client = path_to_curr_module + "/coap_client.py"

    # If destination address is IPv6, check
    # if square brackets are present to use it as
    # a hostname. If not, add square brackets.
    if ipv6:
        if destination_ip[0] != "[":
            destination_ip = "[" + destination_ip + "]"

    return exec_exp_commands(
        f"ip netns exec {ns_name} python3 {path_to_client} -d {destination_ip} {coap_options}",
        stdout=out,
        stderr=err,
    )


def run_coap_server(ns_name, server_options):
    """
    Run CoAP server program

    Parameters
    ----------
    ns_name : str
        Name of the server namespace
    server_options : str
        Options string for running the server process

    Returns
    -------
    int
        return code of the command executed
    """
    # Here, the path to the current file being run `coap.py` is taken
    # and used to run `coap_server.py` in the same directory. This is
    # done using the `__file__` dunder variable and the `os` library.
    path_to_curr_module = os.path.dirname(os.path.abspath(__file__))
    path_to_server = path_to_curr_module + "/coap_server.py"

    return exec_subprocess_in_background(
        f"ip netns exec {ns_name} python3 -u {path_to_server} {server_options}",
        shell=True,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""MPEG DASH commands"""

import os
import tempfile
from .exec import exec_exp_commands, exec_subprocess_in_background


def run_mpeg_dash_http_server(ns_name, port, encoded_chunks_path):
    """
    Run MPEG DASH HTTP server on provided namespace

    Parameters
    ----------
    ns_name : str
        Name of the server namespace
    port: int
        Port number to run the server on.
    encoded_chunks_path : Path
        The path where the encoded chunks are present.

    Returns
    -------
    int
        return code of the command executed
    """

    path_to_server_code = (
        os.path.realpath(os.path.dirname(__file__)) + "/http_server.py"
    )

    cmd_string = f"""ip netns exec {ns_name} \
        python3 {path_to_server_code} {port} {str(encoded_chunks_path)}"""

    return exec_subprocess_in_background(cmd_string)


# pylint: disable=too-many-arguments
def run_mpeg_dash_client(
    ns_name,
    destination_ip,
    port,
    duration,
    out,
    err,
    player,
    enable_audio_playback,
    additional_player_options,
):
    """
    Run MPEG DASH client

    Parameters
    ----------
    ns_name : str
        Name of the client namespace
    destination_ip : str
        IP address of the server namespace
    port : int
        port number of the server at which the MPEG-DASH application is running
    duration :
        Number of seconds for which experiment has to be run
    player: str, optional
        The media player to be used.
    enable_audio_playback: bool
        Enable/disable audio playback
    additional_player_options : string
        User specified options for the video player

    Returns
    -------
    int
        return code of the command executed
    """

    # Obtaining the current linux user's UID
    # in order to configure the PulseAudio sound server
    # to work in a network namespace.

    # pylint: disable=consider-using-with
    try:
        # If sudo has been invoked then the following statements are executed.
        temp_file = tempfile.TemporaryFile()
        # $SUDO_UID is the user ID of the user who invoked sudo
        exec_exp_commands("bash -c 'echo $SUDO_UID' ", stdout=temp_file)
        temp_file.seek(0)
        u_id = int(temp_file.read().decode())
        vlc_binary_prefix = "vlc-wrapper"
    except ValueError:
        # If sudo has not been invoked then the following statements are executed.
        temp_file = tempfile.TemporaryFile()
        exec_exp_commands("bash -c 'echo $UID' ", stdout=temp_file)
        temp_file.seek(0)
        u_id = int(temp_file.read().decode())
        vlc_binary_prefix = "sed -i 's/geteuid/getppid/' /usr/bin/vlc; vlc"

    # Running the media player to stream MPEG-DASH video
    # and setting the quality adaptation logic based on the bandwidth.

    if player == "vlc":
        audio_playback_flag_str = "--no-audio" if not enable_audio_playback else ""

        cmd_string = f"""ip netns exec \
        {ns_name} \
        env PULSE_SERVER=/run/user/{u_id}/pulse/native \
            PULSE_COOKIE=/run/user/{u_id}/pulse/cookie \
        bash -c " \
        {vlc_binary_prefix} \
        http://{destination_ip}:{port}/manifest.mpd \
        --adaptive-logic=rate \
        --loop \
        {audio_playback_flag_str} \
        {additional_player_options}"
        """

    if player == "gpac":
        audio_playback_filter_str = "aout" if enable_audio_playback else ""

        cmd_string = f"""ip netns exec \
        {ns_name} \
        env PULSE_SERVER=/run/user/{u_id}/pulse/native \
            PULSE_COOKIE=/run/user/{u_id}/pulse/cookie \
        bash -c "rm /tmp/gpac_cache -rf; \
        gpac -i \
        http://{destination_ip}:{port}/manifest.mpd"""
        cmd_string += f""":gpac:algo=grate:start_with=min_bw:debug_as=0,1 \
        vout:buffer=1000:mbuffer=5000:cache=none \
        {audio_playback_filter_str} \
        -logs=all@info:ncl \
        -clean-cache \
        -sloop \
        {additional_player_options}"
        """

    return exec_exp_commands(
        cmd_string,
        stdout=out,
        stderr=err,
        timeout=duration,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
NeST internally uses auto-generated unique ids to name topologies.
This module holds the mapping between the user given names and the names
generated internally by NeST.
"""

import json
import logging
from threading import RLock

logger = logging.getLogger(__name__)

DEFAULT_NAMESPACE_ID = "default-namespace-0"

# pylint: disable-msg=R0904


class TopologyMap:
    """
    Store mapping between user given names and NeST ids
    """

    # NOTE:
    # Specifing None in the APIs inplace of ns_id will consider
    # it as adding, deleting or moving device from
    # default namespace, as the API's are compatible to accept None
    # as default namespace

    nodes = {}
    devices = {DEFAULT_NAMESPACE_ID: {}}
    hosts = []
    routers = []
    networks = []
    orphan_interfaces = 0
    lock = RLock()

    @staticmethod
    def add_node(ns_id, node):
        """
        Add node to TopologyMap

        Parameters
        ----------
        ns_id : str
            namespace id
        node : Node
            Object reference of Node class
        """
        # Acquire lock before entering
        with TopologyMap.lock:
            nodes = TopologyMap.get_nodes()

            if ns_id in nodes:
                raise ValueError(f"Node with {ns_id} already exists in the TopologyMap")

            nodes[ns_id] = node
            TopologyMap.devices[ns_id] = {}

    @staticmethod
    def add_host(host):
        """
        Add host to topology_map. This is required apart
        from `add_node` for routing

        Parameters
        ----------
        host : Node
            Object reference of Node class
        """
        with TopologyMap.lock:
            hosts = TopologyMap.get_hosts()

            hosts.append(host)

    @staticmethod
    def add_router(router):
        """
        Add router to topology_map. This is required apart
        from `add_node` for routing

        Parameters
        ----------
        router : Router
            Object reference of Router class
        """
        with TopologyMap.lock:
            routers = TopologyMap.get_routers()
            if router not in routers:
                routers.append(router)
                TopologyMap.get_hosts().remove(
                    router
                )  # remove the router from hosts list

    @staticmethod
    def add_device(ns_id, dev_id, device):
        """
        Adds device to the corresponding node of the TopologyMap

        Parameters
        ----------
        ns_id : str
            node id
        dev_id : str
            device id
        device : Device
            Object reference of Device class
        """
        # Acquire lock before entering
        with TopologyMap.lock:
            devices = TopologyMap.get_devices(ns_id)

            if ns_id is None:
                # Add into default namespace
                devices[dev_id] = device

            else:
                if dev_id in devices:
                    raise ValueError(
                        f"Device with id {dev_id} already present in namespace {ns_id}"
                    )

                devices[dev_id] = device
                if dev_id in TopologyMap.devices[DEFAULT_NAMESPACE_ID]:
                    # remove from default namespace
                    TopologyMap.delete_device(None, dev_id)

                TopologyMap.orphan_interfaces = TopologyMap.orphan_interfaces + 1

    @staticmethod
    def add_network(network):
        """
        Add network object reference to topology_map

        Parameters
        ----------
        network : Network
            Object reference of Network class
        """
        with TopologyMap.lock:
            TopologyMap.networks.append(network)

    @staticmethod
    def get_topology_map():
        """
        Get the entire topology map

        Returns
        -------
        Dict
        """
        with TopologyMap.lock:
            topology_map = {
                "nodes": TopologyMap.nodes,
                "devices": TopologyMap.devices,
                "hosts": TopologyMap.hosts,
                "routers": TopologyMap.routers,
            }

        return topology_map

    @staticmethod
    def get_nodes():
        """
        Get all the nodes of the TopologyMap

        Returns
        ------
        Dict
        """
        with TopologyMap.lock:
            nodes = TopologyMap.nodes

        return nodes

    @staticmethod
    def get_devices(ns_id):
        """
        Get the devices connected to a node

        Parameters
        ----------
        ns_id : str
            node id

        Returns
        -------
        Dict
        """
        with TopologyMap.lock:
            if ns_id is None:
                return TopologyMap.devices[DEFAULT_NAMESPACE_ID]

            if ns_id not in TopologyMap.devices:
                raise ValueError(
                    f"Namespace with id {ns_id} doesn't exist in TopologyMap"
                )

            devices = TopologyMap.devices[ns_id]

        return devices

    @staticmethod
    def get_routers():
        """
        Get all routers added

        Returns
        -------
        List
        """
        with TopologyMap.lock:
            routers = TopologyMap.routers

        return routers

    @staticmethod
    def get_hosts():
        """
        Get all hosts added

        Returns
        -------
        List
        """
        with TopologyMap.lock:
            hosts = TopologyMap.hosts

        return hosts

    @staticmethod
    def get_networks():
        """
        Get all networks

        Returns
        -------
        List
        """
        with TopologyMap.lock:
            networks = TopologyMap.networks

        return networks

    @staticmethod
    def get_node(ns_id):
        """
        Get a node

        Parameters
        ----------
        ns_id : str
            node id

        Returns
        -------
        Node Object
        """
        with TopologyMap.lock:
            if ns_id not in TopologyMap.nodes:
                raise ValueError(
                    f"Namespace with id {ns_id} doesn't exist in TopologyMap"
                )

            node = TopologyMap.nodes[ns_id]

        return node

    @staticmethod
    def get_device(ns_id, dev_id):
        """
        Get a device connected to a node

        Parameters
        ----------
        ns_id : str
            node id
        dev_id : str
            device id

        Returns
        -------
        Device Object
        """
        with TopologyMap.lock:
            if ns_id is None:
                if dev_id not in TopologyMap.devices[DEFAULT_NAMESPACE_ID]:
                    raise ValueError(
                        f"Device with id {dev_id} not present in default namespace"
                    )

                return TopologyMap.devices[DEFAULT_NAMESPACE_ID][dev_id]

            if ns_id not in TopologyMap.nodes:
                raise ValueError(
                    f"Namespace with id {ns_id} doesn't exist in TopologyMap"
                )

            if dev_id not in TopologyMap.devices[ns_id]:
                raise ValueError(
                    f"Device with id {dev_id} doesn't exist in namespace {ns_id}"
                )

            device = TopologyMap.devices[ns_id][dev_id]

        return device

    @staticmethod
    def get_qdiscs(ns_id, dev_id):
        """
        Get qdiscs associated with a device

        Parameters
        ----------
        ns_id : str
            node id
        dev_id : str
            device id

        Returns
        -------
        List
        """
        with TopologyMap.lock:
            device = TopologyMap.get_device(ns_id, dev_id)

        return device.qdisc_list

    @staticmethod
    def delete_all_mapping():
        """
        Delete all mapppings stored in TopologyMap

        """
        with TopologyMap.lock:
            TopologyMap.nodes = {}
            TopologyMap.devices = {DEFAULT_NAMESPACE_ID: {}}
            TopologyMap.hosts = []
            TopologyMap.routers = []
            TopologyMap.networks = []
            TopologyMap.orphan_interfaces = 0

    @staticmethod
    def dump():
        """
        Dump the TopologyMap. (for debugging purpose)

        """
        with TopologyMap.lock:
            topology_map = TopologyMap.get_topology_map()
            logger.debug("Config")
            logger.debug("------")
            logger.debug(json.dumps(topology_map, indent=4))

    @staticmethod
    def decrement_orphan_interfaces():
        """
        Reduce the orphan interface in the topology

        """
        with TopologyMap.lock:
            TopologyMap.orphan_interfaces = TopologyMap.orphan_interfaces - 1

    @staticmethod
    def delete_device(ns_id, dev_id):
        """
        Deletes a device from mentioned namespace

        Parameters
        ----------
        ns_id : str
            node id
        dev_id : str
            Device id which is deleted

        Returns
        -------
        Device Object
        """
        with TopologyMap.lock:
            # Get the device
            device = TopologyMap.get_device(ns_id, dev_id)

            # Delete the device from the namespace
            if ns_id is None:
                del TopologyMap.devices[DEFAULT_NAMESPACE_ID][dev_id]

            else:
                del TopologyMap.devices[ns_id][dev_id]
                TopologyMap.decrement_orphan_interfaces()

            return device

    @staticmethod
    def move_device(src_ns_id, dst_ns_id, dev_id):
        """
        Moves a device from one node to another

        Parameters
        ----------
        src_ns_id : str
            Source node id
        dst_ns_id : str
            Destination node id
        dev_id : str
            Device id which is moved
        """
        with TopologyMap.lock:
            device = TopologyMap.delete_device(src_ns_id, dev_id)
            TopologyMap.add_device(dst_ns_id, dev_id, device)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handle plotting results obtained"""
from pathlib import Path
import matplotlib
from matplotlib import style

# Since plots are not shown in window, use the non-interactive Agg backend.
# Also fixes https://gitlab.com/nitk-nest/nest/-/issues/73
# Many plotting processes were connecting to X-server simultaneously.
# This created errors as X-server couldn't handle requests
matplotlib.use("agg")

# Set plot style
plotter_folder = Path(__file__).resolve().parent
seaborn_paper_config_path = plotter_folder / "seaborn-v0_8-paper.mplstyle"
style.use(seaborn_paper_config_path)
style.use("ggplot")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""
Experiment Module
=================

Handles running network experiments on topologies built using
`Topology` module
"""

from .experiment import (
    Experiment,
    Flow,
    Application,
    CoapApplication,
    MpegDashApplication,
    SipApplication,
)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to device creation in topology"""

import logging
from nest import config
from nest.topology_map import TopologyMap
from nest.topology.id_generator import IdGen
from nest.topology.device.traffic_control import TrafficControlHandler
from nest import engine
import nest.global_variables as g_var
from nest.exception import NestBaseException
from nest.topology.address import Address

logger = logging.getLogger(__name__)

# Max length of device when Topology Map is disabled
# i.e. 'assign_random_names' is set to False in config
MAX_CUSTOM_NAME_LEN = 15

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes


class Device:
    """
    A common subclass which all devices inherit

    Attributes
    ----------

    name : str
        User given name for the device
    id : str
        This value is used by `engine` to create emulated device
        entity
    node_id : str
        This is id of the node that the device belongs to
    traffic_ctl_hanlder : TrafficControlHanlder
        Takes care of traffic control for the device
    """

    def __init__(self, name, node_id):
        """
        Constructore for a device

        Parameters
        ----------
        name : str
            User given name for the device
        node_id : str
            This is the id of the node that the device belongs to
        """

        if config.get_value("assign_random_names") is False:
            if len(name) > MAX_CUSTOM_NAME_LEN:
                raise ValueError(
                    f"Device name {name} is too long. Device names "
                    f"should not exceed 15 characters"
                )
        self._ipv4_address = []
        self._ipv6_address = []
        self._name = name
        self._id = IdGen.get_id(name)
        self._traffic_control_handler = TrafficControlHandler(node_id, self._id)
        if node_id is not None:
            TopologyMap.add_device(self.node_id, self.id, self)

        self._set_structure = False
        self._mtu = 1500
        self._is_mpls_enabled = False

    @property
    def name(self):
        """Getter for name"""
        return self._name

    @property
    def id(self):
        """Getter for id"""
        return self._id

    @property
    def qdisc_list(self):
        """
        Getther for the list of all qdiscs
        """
        return self._traffic_control_handler.qdisc_list

    @property
    def node_id(self):
        """
        Getter for the `Node` associated
        with the device
        """
        return self._traffic_control_handler.node_id

    @node_id.setter
    def node_id(self, node_id):
        """
        Setter for the `Node` associated
        with the device

        Parameters
        ----------
        node : str
            The id of the node where the device is to be installed
        """
        self._traffic_control_handler.node_id = node_id
        if node_id is not None:
            TopologyMap.add_device(self.node_id, self.id, self)

    def set_mode(self, mode):
        """
        Changes the mode of the device

        Parameters
        ----------
        mode : string
            device mode to be set
        """
        if mode in ("UP", "DOWN"):
            if self.node_id is not None:
                engine.set_interface_mode(self.node_id, self.id, mode.lower())
            else:
                # TODO: Create our own error class
                raise NotImplementedError(
                    "You should assign the device to node or router before setting it's mode"
                )
        else:
            raise ValueError(
                f'{mode} is not a valid mode (it has to be either "UP" or "DOWN")'
            )

    def _validate_transform_address(self, func_name, address):
        """
        Internal method to validate Address/str or List[Address/str] input type
        and transform it into List[Address]

        Parameters
        ----------
        func_name : str
            Name of the function that is using this validation
        address : Address/str or List[Address/str]
            IP address(es) to be validated and transformed
        """

        init_type = type(address)
        if isinstance(address, str):
            address = [Address(address)]
        elif isinstance(address, list):
            for i in range(len(address)):
                if isinstance(address[i], str):
                    address[i] = Address(address[i])
                elif not isinstance(address[i], Address):
                    raise TypeError(
                        f"Expected type of argument 'address' in method '{func_name}'"
                        " is Address/str or List[Address/str]. "
                        f"But got input '{address}' of type List[{type(address[i])}]"
                    )
        elif isinstance(address, Address):
            address = [address]
        else:
            raise TypeError(
                f"Expected type of argument 'address' in method '{func_name}'"
                " is Address/str or List[Address/str]. "
                f"But got input '{address}' of type {init_type}"
            )

        if self.node_id is None:
            raise DeviceNotPartOfNodeError(
                "You should assign the interface to node or router before assigning address to it."
            )

        return address

    def set_address(self, address):
        """
        Assigns IP address/addresses to an interface

        Parameters
        ----------
        address : Address/str or List[Address/str]
            IP address to be assigned to the interface
        """

        # Make arguments to a list of Address objects and validate type
        address = self._validate_transform_address("set_address", address)

        # Deleting old ip addresses
        for addr in self._ipv4_address:
            engine.delete_ip(self.node_id, self.id, addr.get_addr())
        self._ipv4_address = []
        for addr in self._ipv6_address:
            engine.delete_ip(self.node_id, self.id, addr.get_addr())
        self._ipv6_address = []

        # Adding new ip addresses
        for addr in address:
            engine.assign_ip(self.node_id, self.id, addr.get_addr())
            if addr.is_ipv6() is True:
                self._ipv6_address.append(addr)
            else:
                self._ipv4_address.append(addr)

        # Global variable to check if address is ipv6 or not for DAD check
        if len(self._ipv6_address) != 0:
            g_var.IS_IPV6 = True

    def add_address(self, address):
        """
        Adds IP address to an interface

        Parameters
        ----------
        address : Address/str or List[Address/str]
            IP address to be added to the interface
        """

        # Make arguments to a list of Address objects and validate type
        address = self._validate_transform_address("add_address", address)

        for addr in address:
            engine.assign_ip(self.node_id, self.id, addr.get_addr())
            if addr.is_ipv6() is True:
                self._ipv6_address.append(addr)
            else:
                self._ipv4_address.append(addr)

    def get_address(self, ipv4=True, ipv6=True, as_list=False):
        """
        Gets the required IP addresses for the interface
        Returns a list or an Address object
        Parameters
        ----------
        ipv4 : If set to true, the IPv4 address of the interface is returned (defaults to True)
        ipv6 : If set to true, the IPv6 address of the interface is returned (defaults to True)
        If both are True, both the addresses are returned
        Either ipv4 or ipv6 must be True
        as_list : Only applicable when a single address is set (Applicable individually for
                  ipv4 and ipv6).
        Returns Address object when false, else returns a single Address object in list
        (defaults to False).
        """

        # check if one of ipv4/ipv6 is true
        if not (ipv4 or ipv6):
            raise ValueError("Either ipv4 or ipv6 argument must be True")

        list_of_addresses = []

        if ipv4:
            list_of_addresses += self._ipv4_address
        if ipv6:
            list_of_addresses += self._ipv6_address

        if len(list_of_addresses) == 1 and as_list is False:
            return list_of_addresses[0]

        return list_of_addresses

    def del_address(self, address):
        """
        Delete IP address(es) from the interface

        Parameters
        ----------
        address : str or list
            IP address to be deleted from the interface
        """
        # Make arguments to a list of Address objects and validate type
        address = self._validate_transform_address("del_address", address)

        for addr in address:
            deleted = None
            if addr.is_ipv6():
                for i in range(len(self._ipv6_address)):
                    if self._ipv6_address[i].get_addr() == addr.get_addr():
                        engine.delete_ip(self.node_id, self.id, addr.get_addr())
                        deleted = self._ipv6_address.pop(i)
                        break
            else:
                for i in range(len(self._ipv4_address)):
                    if self._ipv4_address[i].get_addr() == addr.get_addr():
                        engine.delete_ip(self.node_id, self.id, addr.get_addr())
                        deleted = self._ipv4_address.pop(i)
                        break
            if deleted is None:
                addrs = addr.get_addr()
                logger.warning(
                    "Cannot delete an address that is not assigned. (%s)", addrs
                )

    def add_qdisc(self, qdisc, parent="root", handle="", **kwargs):
        """
        Add a qdisc (Queueing Discipline) to this device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        dev : Interface class
            The device to which the qdisc is to be added
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        """

        self._traffic_control_handler.add_qdisc(qdisc, parent, handle, **kwargs)

    def change_qdisc(self, handle, qdisc="", **kwargs):
        """
        Change a qdisc that is already present in the device

        Parameters
        ----------
        handle : string
            Hande of the qdisc to be changed
        qdisc : string
            The new qdisc to be added to the device
        """

        self._traffic_control_handler.change_qdisc(handle, qdisc, **kwargs)

    def delete_qdisc(self, handle):
        """
        Delete qdisc (Queueing Discipline) from this device

        Parameters
        ----------
        handle : string
            Handle of the qdisc to be deleted
        """

        self._traffic_control_handler.delete_qdisc(handle)

    def add_class(self, qdisc, parent="root", classid="", **kwargs):
        """
        Create an object that represents a class

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """

        self._traffic_control_handler.add_class(qdisc, parent, classid, **kwargs)

    def change_class(self, qdisc, parent, classid, **kwargs):
        """
        Change a class that is already present in the device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """

        self._traffic_control_handler.change_class(qdisc, parent, classid, **kwargs)

    def delete_class(self, classid, parent):
        """
        Delete Class from this device

        Parameters
        ----------
        handle : string
            Handle of the class to be deleted
        """
        self._traffic_control_handler.delete_class(classid, parent)

    # pylint: disable=too-many-arguments
    def add_filter(
        self,
        dev_id,
        priority,
        filtertype,
        flowid,
        protocol="ip",
        parent="root",
        handle="",
        **kwargs,
    ):
        """
        Design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        dev_id : string
            Device from the packets are being mirrored
        protocol : string
            protocol used (Default value = 'ip')
        priority : int
            priority of the filter
        filtertype : string
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        filter : dictionary
            filter parameters
        """
        # TODO: Verify type of parameters
        # TODO: Reduce arguments to the engine functions by finding parent and handle automatically

        self._traffic_control_handler.add_filter(
            dev_id, priority, filtertype, flowid, protocol, parent, handle, **kwargs
        )

    def delete_filter(self, handle, parent):
        """
        Delete filter from a Class or a Qdisc

        Parameters
        ----------
        handle : string
            Handle of the class to be deleted
        """

        self._traffic_control_handler.delete_filter(handle, parent)

    def enable_offload(self, offload_name):
        """
        API for enabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to enable
        """
        if not isinstance(offload_name, list):
            offload_name = [offload_name]
        valid_offloads(offload_name)
        namespace_id = self.node_id
        interface_id = self.id
        for offload_type in offload_name:
            if engine.ethtool.enable_offloads(namespace_id, interface_id, offload_type):
                logger.debug(
                    "%s is enabled on interface %s",
                    offload_type,
                    self.name,
                )
            else:
                logger.error(
                    "%s is not enabled on interface %s",
                    offload_type,
                    self.name,
                )

    def disable_offload(self, offload_name):
        """
        API for disabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to disable
        """
        if not isinstance(offload_name, list):
            offload_name = [offload_name]
        valid_offloads(offload_name)
        namespace_id = self.node_id
        interface_id = self.id
        for offload_type in offload_name:
            if engine.ethtool.disable_offloads(
                namespace_id, interface_id, offload_type
            ):
                logger.debug(
                    "%s is disabled on interface %s",
                    offload_type,
                    self.name,
                )
            else:
                logger.error(
                    "%s is not disabled on interface %s",
                    offload_type,
                    self.name,
                )

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for an interface.
        """
        engine.disable_dad(self.node_id, self._id)

    def enable_mpls(self):
        """
        Enables mpls input through the interface.
        Requires mpls kernel modules to be loaded.

        Run ``sudo modprobe mpls_iptunnel`` to load mpls modules.
        """
        if self.node_id is None:
            raise NotImplementedError(
                "You should assign the interface to node or router before enabling mpls"
            )

        engine.set_mpls_max_label_node(self.node_id, int(100000))

        if self._is_mpls_enabled is False:
            engine.enable_mpls_interface(self.node_id, self.id)
            self._is_mpls_enabled = True
            self.mtu = 1504

    def is_mpls_enabled(self):
        """
        Check if the interface has mpls enabled
        """
        return self._is_mpls_enabled

    def enable_mptcp_endpoint(self, flags):
        """
        Declares the interface as an MPTCP endpoint
        """
        engine.enable_mptcp_endpoint(self.node_id, self, flags)

    def get_mptcp_endpoints(self):
        """
        Gets the MPTCP endpoints of the interface
        """
        return engine.get_mptcp_endpoints(self.node_id, self)

    @property
    def mtu(self):
        """
        Get the maximum transmit unit value for the interface
        """
        return self._mtu

    @mtu.setter
    def mtu(self, mtu_value):
        """
        Set the maximum transmit unit value for the interface
        Default is 1500 bytes.
        """
        if self._mtu != mtu_value:
            engine.set_mtu_interface(self.node_id, self.id, int(mtu_value))
            self._mtu = mtu_value
            logger.debug("MTU of interface %s set to %s", self.name, str(self.mtu))

    def set_structure(self):
        """
        Sets a proper structure to the interface by creating HTB class
        with default bandwidth and a netem qdisc as a child.
        HTB is used for setting bandwidth and netem is used for setting
        delay, packet corruption etc
        (default bandwidth = 1024mbit)
        """

        if self._set_structure is True:
            return

        self._set_structure = True

        default_route = {"default": "1"}

        # HTB is added since netem is a classless qdisc. So, htb class,
        # With netem as child is added
        self.add_qdisc("htb", "root", "1:", **default_route)

        # TODO: find how to set a good bandwidth
        default_bandwidth = {"rate": config.get_value("default_bandwidth")}

        self.add_class("htb", "1:", "1:1", **default_bandwidth)

        self.add_qdisc("netem", "1:1", "11:")

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"


def valid_offloads(offload_name):
    """
    Check valid offloads

    Parameters
    -----------
    offload_name : str
        The offload name
    """
    offloads_list = ["tso", "gso", "gro"]
    for offload_type in offload_name:
        if not offload_type in offloads_list:
            logger.error("Invalid offload")
            raise ValueError(f"{offload_type} is not a valid offload")


class DeviceNotPartOfNodeError(NestBaseException):
    """
    Exception related to error in not assigning Device to Node.
    """



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""User API to setup and run experiments on a given topology"""

import copy
import logging
from collections import defaultdict
from pathlib import Path
import random
import sys
from nest import engine, config
from nest.input_validator.metric import Bandwidth
from nest.network_utilities import ipv6_dad_check
from nest.input_validator import input_validator
from nest.topology import Node, Address
from nest.topology.interface import BaseInterface
from nest.topology_map import TopologyMap
from .run_exp import run_experiment
from .pack import Pack
from .tools import Iperf3Options
from .helpers.tcp_validations import TCPValidations

logger = logging.getLogger(__name__)
tcp_validator = TCPValidations()


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class Flow:
    """Defines a flow in the topology"""

    # pylint: disable=too-many-arguments
    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        destination_address: Address,
        start_time: int,
        stop_time: int,
        number_of_streams: int,
        source_address: Address = None,
    ):
        """
        'Flow' object in the topology

        Parameters
        ----------
        source_node : Node
            Source node of flow
        destination_node : Node
            Destination node of flow
        destination_address : Address/str
            Destination address of flow
        start_time : int
            Time to start flow (in seconds)
        stop_time : int
            Time to stop flow (in seconds)
        number_of_streams : int
            Number of streams in the flow
        source_address : Address/str
            Source address of flow
        """
        self.source_node = source_node
        self.destination_node = destination_node
        self.destination_address = destination_address
        self.start_time = start_time
        self.stop_time = stop_time
        self.number_of_streams = number_of_streams
        if source_address is None:
            self.source_address = source_node.interfaces[0].get_address()
        else:
            self.source_address = source_address

        self._options = {"protocol": "TCP", "cong_algo": "cubic"}
        self.user_input_options = {}

    @property
    def source_address(self):
        """Getter for source address"""
        return self._source_address

    @source_address.setter
    def source_address(self, source_address):
        """Setter for source address"""
        if isinstance(source_address, str):
            source_address = Address(source_address)
        self._source_address = source_address

    @property
    def destination_address(self):
        """Getter for destination address"""
        return self._destination_address

    @destination_address.setter
    def destination_address(self, destination_address):
        """Setter for destination address"""
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)
        self._destination_address = destination_address

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """

        return [
            self.source_node.id,
            self.destination_node.id,
            self.destination_address.get_addr(with_subnet=False),
            self.start_time,
            self.stop_time,
            self.number_of_streams,
            self._options,
            self.source_address.get_addr(with_subnet=False),
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r}, {self.source_address!r},"
            f" {self.destination_address!r}), {self.start_time!r}, {self.stop_time!r}"
            f" {self.number_of_streams!r})"
        )

    @property
    def protocol(self):
        """Getter for protocol"""
        return self._options["protocol"]

    @staticmethod
    @input_validator
    # pylint: disable=too-many-arguments
    def setup_mptcp_connection(
        source_interface: BaseInterface,
        destination_interface: BaseInterface,
        start_time: int,
        stop_time: int,
        number_of_streams: int = 1,
    ):
        """
        Creates and returns an MPTCP flow for given source and destination
        interface with all MPTCP endpoints enabled on both nodes

        Parameters
        ----------
        source_interface : Interface
            Source interface for MPTCP Flow
        destination_interface : Interface
            Destination interface for MPTCP Flow
        start_time : int
            Start time for the flow
        stop_time : int
            Stop time for the flow
        number_of_streams : int = 1
            Number of streams in the flow

        Returns
        -------
        Flow
            An MPTCP capable flow from source_interface to destination_interface
        """

        # Enable MPTCP on the Source Nodes
        source_node = TopologyMap.get_node(source_interface.node_id)
        source_node.enable_mptcp()

        # Enable MPTCP on the Destination Node
        destination_node = TopologyMap.get_node(destination_interface.node_id)
        destination_node.enable_mptcp()

        # Set MPTCP limits on both the nodes
        add_addr_limit = min(
            5, max(len(source_node.interfaces), len(destination_node.interfaces))
        )
        max_subflow_limit = min(
            5, len(source_node.interfaces) * len(destination_node.interfaces) - 1
        )
        source_node.set_mptcp_parameters(
            max_subflows=max_subflow_limit, max_add_addr_accepted=add_addr_limit
        )
        destination_node.set_mptcp_parameters(
            max_subflows=max_subflow_limit, max_add_addr_accepted=add_addr_limit
        )

        # Check if MPTCP can be run between the 2 hosts as per RFC
        if not (
            source_node.is_mptcp_supported() or destination_node.is_mptcp_supported()
        ):
            logger.error(
                "Neither of your source node and destination node is multihomed and "
                "multiadressed. Please check your topology and try again."
            )
            sys.exit()

        # If MPTCP can be run, enable all interfaces as MPTCP endpoints with default flags
        for interface in source_node.interfaces:
            interface.enable_mptcp_endpoint(
                ["subflow", "fullmesh"]
                if "fullmesh" in engine.possible_mptcp_flags
                else ["subflow"]
            )
        for interface in destination_node.interfaces:
            interface.enable_mptcp_endpoint(["signal"])

        # Create and return a flow that will work with MPTCP
        return Flow(
            source_node,
            destination_node,
            destination_interface.get_address(),
            start_time,
            stop_time,
            number_of_streams,
            source_address=source_interface.get_address(),
        )


class Application:
    """Defines an application in the topology"""

    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        destination_address: Address,
    ):
        """
        'Application' object in the topology

        Parameters
        ----------
        source_node : Node
            Source node of application
        destination_node : Node
            Destination node of application
        destination_address : Address/str
            Destination address of application
        """
        self.source_node = source_node
        self.destination_node = destination_node
        self.destination_address = destination_address

    @property
    def destination_address(self):
        """Getter for destination address"""
        return self._destination_address

    @destination_address.setter
    def destination_address(self, destination_address):
        """Setter for destination address"""
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)
        self._destination_address = destination_address

    def _get_props(self):
        """
        Get application properties.

        NOTE: To be used internally
        """

        return [
            self.source_node.id,
            self.destination_node.id,
            self.destination_address.get_addr(with_subnet=False),
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r},"
            f" {self.destination_address!r})"
        )


class CoapApplication(Application):
    """Defines a CoAP application in the topology"""

    # pylint: disable=too-many-arguments
    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        destination_address: Address,
        n_con_msgs: int,
        n_non_msgs: int,
        user_options=None,
    ):
        """
        Application object representing CoAP applications in the topology.
        Inherited from the `Application` class.
        Application object representing CoAP applications in the topology.
        Inherited from the `Application` class.

        Parameters
        ----------
        source_node : Node
            Source node of flow
        destination_node : Node
            Destination node of flow
        destination_address : Address/str
            Destination address of flow
        n_con_msgs : int
            Number of confimable messages to be sent in the flow
        n_non_msgs : int
            Number of non-confimable messages to be sent in the flow
        user_options : dict, optional
            User specified options for particular tools
        """
        self.n_con_msgs = n_con_msgs
        self.n_non_msgs = n_non_msgs

        # Options for users to set
        self.user_options = user_options
        super().__init__(source_node, destination_node, destination_address)
        super().__init__(source_node, destination_node, destination_address)

    # Destination address getter and setter are implemented
    # in the Application class which is the superclass of CoapApplication class
    # in the Application class which is the superclass of CoapApplication class

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """

        return [
            self.source_node.id,
            self.destination_node.id,
            self.destination_address.get_addr(with_subnet=False),
            self.n_con_msgs,
            self.n_non_msgs,
            self.user_options,
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r},"
            f" {self.destination_address!r}),"
            f" {self.n_con_msgs!r}, {self.n_non_msgs!r}, {self.user_options!r})"
        )


class MpegDashApplication(Application):
    """Defines an MPEG DASH application in the topology"""

    supported_media_players = ["gpac", "vlc"]

    # pylint: disable=too-many-arguments
    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        source_address: Address,
        destination_address: Address,
        port: int,
        encoded_chunks_path: Path,
        duration: int,
        player: str,
        enable_audio_playback: bool = False,
        additional_player_options=None,
    ):
        """
        Application object representing MPEG-DASH applications in the topology.
        Inherited from the `Application` class.

        Parameters
        ----------
        source_node : Node
            Node that behaves as MPEG-DASH client
        destination_node : Node
            Node that behaves as MPEG-DASH server
        source_address : Address/str
            Address of the MPEG-DASH client
        destination_address : Address/str
            Address of the MPEG-DASH server
        port : int
            Port number of the server at which MPEG-DASH application is running
        encoded_chunks_path : Path
            The path where the encoded chunks are stored.
        duration : int
            Number of seconds for which experiment has to be run
        player: str
            The media player to be used.
        enable_audio_playback: bool, Default is False (i.e. no audio)
            Enable/disable audio playback
        additional_player_options : list, optional
            User specified options for the video player
        """
        self.source_address = source_address
        self.port = port
        self.encoded_chunks_path = encoded_chunks_path
        self.duration = duration
        self.player = player
        self.enable_audio_playback = enable_audio_playback
        self.additional_player_options = additional_player_options

        super().__init__(source_node, destination_node, destination_address)

        if self.player not in self.supported_media_players:
            raise RuntimeError(
                f"Invalid media player {self.player}!",
                "NeST currently supports the following media players:",
                f"{self.supported_media_players}.",
                "Please specify one of these.",
            )

        if self.additional_player_options is not None:
            self.additional_player_options = " ".join(self.additional_player_options)
        else:
            self.additional_player_options = ""

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """
        return [
            self.source_node.id,
            self.destination_node.id,
            self.source_address.get_addr(with_subnet=False),
            self.destination_address.get_addr(with_subnet=False),
            self.port,
            self.encoded_chunks_path,
            self.duration,
            self.player,
            self.enable_audio_playback,
            self.additional_player_options,
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r},"
            f" {self.source_address!r}, {self.destination_address!r}, {self.port!r},"
            f" {self.encoded_chunks_path!r}, {self.duration!r},"
            f" {self.player!r}, {self.enable_audio_playback !r} {self.additional_player_options!r})"
        )


class SipApplication(Application):
    """Defines a SIP application in the topology"""

    # pylint: disable=too-many-arguments
    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        source_address: Address,
        destination_address: Address,
        port: int,
        duration: int,
        scenario,
        server_xml=None,
        client_xml=None,
        callrate=None,
    ):
        """
        Application object representing SIP applications in the topology.
        Inherited from the `Application` class.

        Parameters
        ----------
        source_node : Node
            Node that behaves as SIP client
        destination_node : Node
            Node that behaves as SIP server
        source_address : Address/str
            Address of the SIP client
        destination_address : Address/str
            Address of the SIP server
        port : int
            Port number of the server at which SIP application is running
        duration:
            Duration of experiment
        scenario:
            The integrated scenario to run.
            ("basic", "basic_with_audio", "branch" or "xml")
        server_xml:
            Custom XML file for running sipp server instance,
            should be passed only when scenario is "xml"
        client_xml:
            Custom XML file for running sipp client instance,
            should be passed only when scenario is "xml"
        callrate:
            Number of calls client will try per second (default is 10)

        """
        self.source_address = source_address
        self.port = port
        self.duration = duration
        self.scenario = scenario
        self.server_xml = server_xml
        self.client_xml = client_xml
        self.callrate = callrate

        super().__init__(source_node, destination_node, destination_address)

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """
        return [
            self.source_node.id,
            self.destination_node.id,
            self.source_address.get_addr(with_subnet=False),
            self.destination_address.get_addr(with_subnet=False),
            self.port,
            self.duration,
            self.scenario,
            self.server_xml,
            self.client_xml,
            self.callrate,
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r},"
            f" {self.source_address!r}, {self.destination_address!r},"
            f" {self.port!r}, {self.duration!r}, {self.scenario!r},"
            f" {self.server_xml!r}, {self.client_xml!r}, {self.callrate!r}"
        )


class Experiment:
    """Handles experiment to be run on topology"""

    # Stores configuration of old and new congestion algorithms for cleanup
    old_cong_algos = defaultdict(dict)
    new_cong_algos = []

    @input_validator
    def __init__(self, name: str):
        """
        Create experiment

        Parameters
        ----------
        name : str
            Name of experiment
        """
        self.name = name
        self.flows = []
        self.coap_applications = []
        self.mpeg_dash_applications = []
        self.sip_applications = []
        self.node_stats = []
        self.qdisc_stats = []
        self.tcp_module_params = defaultdict(dict)

    def add_flow(self, flow):
        """
        Add flow to experiment
        By default, the flow is assumed to be
        TCP with cubic congestion algorithm

        Parameters
        ----------
        flow : Flow
            Add flow to experiment
        """
        self.flows.append(copy.deepcopy(flow))

    @input_validator
    def add_tcp_flow(
        self,
        flow: Flow,
        congestion_algorithm="cubic",
        tool: str = "netperf",
        **kwargs: dict,
    ):
        """
        Add TCP flow to experiment. If no congestion control algorithm
        is specified, then by default cubic is used.

        Note: The congestion control algorithm specified in this API
        overrides the congestion control algorithm specified in
        `topology.Node.configure_tcp_param()` API.

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        congestion_algorithm : str
            TCP congestion algorithm (Default value = 'cubic')
        tool : str
            Tool (netperf/iperf3) to be used for the experiment (Default value = 'netperf')
        """

        tcp_validator.verify_tool(tool)
        tcp_validator.verify_congestion_control_algorithm(congestion_algorithm)

        # TODO: Verify congestion algorithm

        options = {
            "protocol": "TCP",
            "cong_algo": congestion_algorithm,
            "tool": tool,
        }

        if tool == "iperf3":
            user_options = {}

            for params, value in kwargs.items():
                if params == "server_options":
                    user_options.update(value)

                if params == "client_options":
                    user_options.update(value)

            iperf3options = Iperf3Options(kwargs=user_options).getter()
            if (
                "port_nos" not in iperf3options
                or len(iperf3options["port_nos"]) < flow.number_of_streams
            ):
                port_nos = set()
                # If user has provided certain port_nos, use them
                if iperf3options.get("port_nos") is not None:
                    port_nos = set(iperf3options["port_nos"])
                while len(port_nos) < flow.number_of_streams:
                    port_nos.add(random.randrange(1024, 65536))
                iperf3options.update({"port_nos": list(port_nos)})

            # iperf3options.update(options)
            options = {**iperf3options, **options}

        flow._options = options  # pylint: disable=protected-access
        self.add_flow(flow)

    @input_validator
    def add_mptcp_flow(
        self,
        flow: Flow,
        congestion_algorithm="cubic",
        tool: str = "netperf",
        **kwargs: dict,
    ):
        """
        Add MPTCP flow to experiment. If no congestion control algorithm
        is specified, then by default cubic is used.

        Note: The congestion control algorithm specified in this API
        overrides the congestion control algorithm specified in
        `topology.Node.configure_tcp_param()` API.

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        congestion_algorithm : str
            TCP congestion algorithm (Default value = 'cubic')
        tool : str
            Tool (netperf/iperf3) to be used for the experiment
            (Default value = 'netperf')
        """
        tcp_validator.verify_tool(tool)
        tcp_validator.verify_congestion_control_algorithm(congestion_algorithm)

        options = {
            "protocol": "MPTCP",
            "cong_algo": congestion_algorithm,
            "tool": tool,
        }

        if tool == "iperf3":
            user_options = {}

            for params, value in kwargs.items():
                if params == "server_options":
                    user_options.update(value)

                if params == "client_options":
                    user_options.update(value)

            iperf3options = Iperf3Options(kwargs=user_options).getter()
            if (
                "port_nos" not in iperf3options
                or len(iperf3options["port_nos"]) < flow.number_of_streams
            ):
                port_nos = set()
                # If user has provided certain port_nos, use them
                if iperf3options.get("port_nos") is not None:
                    port_nos = set(iperf3options["port_nos"])
                while len(port_nos) < flow.number_of_streams:
                    port_nos.add(random.randrange(1024, 65536))
                iperf3options.update({"port_nos": list(port_nos)})

            # iperf3options.update(options)
            options = {**iperf3options, **options}

        flow._options = options  # pylint: disable=protected-access
        self.add_flow(flow)

    @input_validator
    def add_udp_flow(
        self,
        flow: Flow,
        target_bandwidth: Bandwidth = Bandwidth("1mbit"),
        server_options: dict = None,
        client_options: dict = None,
    ):
        """
        Add UDP flow to experiment

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        target_bandwidth :
            UDP bandwidth (in Mbits) (Default value = '1mbit')
            This bandwidth limit is for each UDP stream in the flow
        """
        options = {"protocol": "udp", "target_bw": target_bandwidth.string_value}

        # options update with user configuration
        user_options = {}
        if server_options:
            user_options.update(server_options)
        if client_options:
            user_options.update(client_options)

        iperf3options = Iperf3Options(protocol="udp", kwargs=user_options).getter()

        if "port_no" not in iperf3options:
            iperf3options.update({"port_no": random.randrange(1024, 65536)})

        iperf3options.update(options)

        # pylint: disable=protected-access
        flow._options = iperf3options
        self.add_flow(flow)

    @input_validator
    def add_coap_application(self, coap_application: CoapApplication):
        """
        Add a CoAP application to experiment

        Parameters
        ----------
        coap_application : CoapApplication
            The coap application to be added to experiment
        """
        self.coap_applications.append(copy.deepcopy(coap_application))

    @input_validator
    def add_mpeg_dash_application(self, mpeg_dash_applications):
        """
        Add an MPEG-DASH application to the experiment

        Parameters
        ----------
        mpeg_dash_applications : Union[MpegDashApplication,list]
            The MPEG-DASH application(s) to be added to experiment
        """
        if isinstance(mpeg_dash_applications, MpegDashApplication):
            self.mpeg_dash_applications.append(copy.deepcopy(mpeg_dash_applications))

        elif isinstance(mpeg_dash_applications, list):
            self.mpeg_dash_applications.extend(copy.deepcopy(mpeg_dash_applications))

        else:
            raise Exception(
                """The add_mpeg_dash_application function takes either
                a MpegDashApplication object or a list of the same only."""
            )

    @input_validator
    def add_sip_application(self, sip_applications):
        """
        Add a SIP application to the experiment

        Parameters
        ----------
        sip_applications : Union[SipApplication,list]
            The SIP application(s) to be added to experiment
        """
        if isinstance(sip_applications, SipApplication):
            self.sip_applications.append(copy.deepcopy(sip_applications))
        elif isinstance(sip_applications, list):
            self.sip_applications.extend(copy.deepcopy(sip_applications))
        else:
            raise Exception(
                """The add_sip_application function takes either
                a SipApplication object or a list of the same only."""
            )

    @input_validator
    def require_qdisc_stats(self, interface: BaseInterface, stats=""):
        """
        Stats to be obtained from qdisc in interface

        Parameters
        ----------
        interface : BaseInterface
            Interface containing the qdisc
        stats : list(str)
            Stats required (Default value = '') [NOT SUPPORTED]
        """
        # TODO: Leads to rewrite if the function is called
        # twice with same 'interface'

        # for stat in stats:
        #     if stat not in Experiment.qdisc_stats:
        #         raise ValueError('{} is not a valid Queue property.'.format(stat))

        if interface.get_qdisc() is None:
            raise ValueError("Given interface hasn't been assigned any qdisc.")

        self.qdisc_stats.append(
            {
                "ns_id": interface.node_id,
                "int_id": interface.ifb_id,
                "qdisc": interface.get_qdisc().qdisc,
                "stats": stats,
            }
        )

    def configure_tcp_module_params(self, congestion_algorithm, **kwargs):
        """
        Set TCP module parameters

        Parameters
        ----------
        congestion_algorithm : str
            TCP congestion algorithm
        **kwargs :
            module parameters to set
        """
        self.tcp_module_params[congestion_algorithm].update(kwargs)
        logger.info("TCP module parameters will be set when the experiment is run")

    @ipv6_dad_check
    def run(self):
        """Run the experiment"""
        print()
        logger.info("Running experiment %s ", self.name)
        Pack.init(self.name)
        if config.get_value("show_mptcp_checklist"):
            tcp_validator.verify_mptcp_setup(self)
        run_experiment(self)

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Test topology map API"""

import unittest
from nest.topology.device.device import Device
from nest.topology_map import TopologyMap
from nest.topology import Node, connect

# pylint: disable=missing-docstring
# pylint: disable=too-many-instance-attributes


class TestTopologyMap(unittest.TestCase):
    def setUp(self):
        # Create nodes
        ns1 = Node("n1")
        ns2 = Node("n2")

        # Get their corresponding ids and names
        self.ns_id1 = ns1.id
        self.ns_name1 = ns1.name
        self.ns_id2 = ns2.id
        self.ns_name2 = ns2.name

        # Create veth pairs (interfaces)
        (eth1, eth2) = connect(ns1, ns2)

        # Get their corresponding ids and names
        self.int_id1 = eth1.id
        self.int_name1 = eth1.name
        self.int_id2 = eth2.id
        self.int_name2 = eth2.name

    def tearDown(self):
        TopologyMap.delete_all_mapping()

    def test_add_and_get_node(self):
        self.assertEqual(TopologyMap.get_node(self.ns_id1).name, self.ns_name1)
        self.assertEqual(TopologyMap.get_node(self.ns_id2).name, self.ns_name2)

    def test_add_and_get_device(self):
        self.assertEqual(
            TopologyMap.get_device(self.ns_id1, self.int_id1).name, self.int_name1
        )
        self.assertEqual(
            TopologyMap.get_device(self.ns_id2, self.int_id2).name, self.int_name2
        )

    def test_add_same_entity_again(self):
        dummy_node = Node("nd")
        dummy_device = Device("device", None)

        with self.assertRaises(ValueError):
            TopologyMap.add_node(self.ns_id1, dummy_node)
        with self.assertRaises(ValueError):
            TopologyMap.add_device(self.ns_id1, self.int_id1, dummy_device)

    def test_delete_device(self):
        device = TopologyMap.get_device(self.ns_id1, self.int_id1)

        TopologyMap.delete_device(self.ns_id1, self.int_id1)
        with self.assertRaises(ValueError):
            TopologyMap.get_device(self.ns_id1, self.int_id1)

        TopologyMap.add_device(self.ns_id1, self.int_id1, device)
        self.assertEqual(
            TopologyMap.get_device(self.ns_id1, self.int_id1).name, self.int_name1
        )

    def test_move_device(self):
        TopologyMap.move_device(self.ns_id1, self.ns_id2, self.int_id1)

        with self.assertRaises(ValueError):
            TopologyMap.get_device(self.ns_id1, self.int_id1)

        self.assertEqual(
            TopologyMap.get_device(self.ns_id2, self.int_id1).name, self.int_name1
        )

        TopologyMap.move_device(self.ns_id2, self.ns_id1, self.int_id1)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Execute Linux commands"""

import logging
import shlex
import subprocess
from subprocess import Popen, PIPE


logger = logging.getLogger(__name__)

# pylint: disable=inconsistent-return-statements
def exec_subprocess(cmd, shell=False, output=False):
    """
    Executes a command

    Parameters
    ----------
    cmd : str
        command to be executed
    shell : boolean
        Spawns a shell and executes the command if true
        (Default value = False)
    output : boolean
        True if the output of the `cmd` is to be returned
        (Default value = False)

    Returns
    -------
    int
        Return code received after executing the command
    """

    # TODO: Commands with pipes are easily executed when
    # they are run within a shell.
    # But it may not be the most efficient way.

    temp_cmd = cmd
    if shell is False:
        temp_cmd = cmd.split()

    with Popen(temp_cmd, stdout=PIPE, stderr=PIPE, shell=shell) as proc:

        (stdout, _) = proc.communicate()
        logger.trace(cmd)

        if output:
            return stdout.decode()
        return proc.returncode


def exec_subprocess_with_live_output(cmd):
    """
    Executes command and prints live output to stdout.

    For eg., if ping command is used, then the output for each
    packet is printed live to stdout.

    Parameters
    ----------
    cmd : str
        Command to be executed

    Returns
    -------
    int
        Return code recieved after executing the command
    """
    # Inspired from:
    # https://fabianlee.org/2019/09/15/python-getting-live-output-from-subprocess-using-poll/

    with Popen(cmd.split(), stdout=PIPE) as process:
        while True:
            output = process.stdout.readline()
            if process.poll() is not None:
                break
            if output:
                print(output.decode(), end="")

        # Print an extra newline at the end
        print()
        logger.trace(cmd)
        return process.poll()


def exec_exp_commands(
    cmd, stdout=PIPE, stderr=PIPE, timeout=None, signal_on_timeout=None
):
    """
    executes experiment related commands like ss, tc and netperf

    Parameters
    ----------
    cmd : str
        command to be executed
    stdout : File
        temp file(usually) to store the output (Default value = subprocess.PIPE)
    stderr : File
        temp file(usually) to store errors, if any (Default value = subprocess.PIPE)
    timeout :
         (Default value = None)
    signal_on_timeout : Signal
        signal to send to process on timeout. (Default = None)
        SIGKILL is sent by default if signal_on_timeout is not passed
    Returns
    -------
    int
        Return code recieved after executing the command
    """
    with Popen(shlex.split(cmd), stdout=stdout, stderr=stderr) as proc:
        logger.trace(cmd)
        try:
            proc.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            if signal_on_timeout is not None:
                proc.send_signal(signal_on_timeout)

                # wait for the process to terminate so that,
                # proc.returncode is set to return value
                proc.wait()
            else:
                proc.kill()
            stderr.write(b"Connection timeout")
        return proc.returncode


def exec_subprocess_in_background(cmd, shell=False, wait_for_exit_code=False):
    """
    Executes a command in the background. This is mainly for
    commands or programs that have an event loop running,
    like server components of networking tools, that need to be
    run in the background so that NeST is uninterrupted.

    If a custom program is to be run using the command,
    the `wait_for_exit_code` parameter can be set to true,
    requiring the custom program to not print anything to
    stdout except the string 'Process successful\n' when
    it can be confirmed that the process has run successfully
    till the point of the event loop. This signifies a success
    in running the process.

    If `wait_for_exit_code` is set to False, the function always
    returns 0 as the exit code, indicating success. This is used
    in situations where the above condition of printing only
    'Process successful\n' to stdout is not possible.

    Parameters
    ----------
    cmd : str
        command to be executed
    shell : boolean
        Spawns a shell and executes the command if true
        (Default value = False)
    wait_for_exit_code: boolean
        To wait or not to wait for a success message from
        the subprocess

    Returns
    -------
    int
        Return code received after executing the command
    """
    # If shell is set to True, use the command as it is
    temp_cmd = cmd

    # If shell is set to False, split the command
    if shell is False:
        temp_cmd = cmd.split()

    # Create and run the subprocess
    # pylint:disable=consider-using-with
    proc = Popen(temp_cmd, stdout=PIPE, stderr=PIPE, shell=shell)

    # If wait_for_exit_code is set to True
    if wait_for_exit_code:
        # Continuously read from the stdout buffer till newline
        # is encountered
        line_in_bytes = proc.stdout.readline()
        line_str = line_in_bytes.decode("utf-8")
        if line_str == "Process successful\n":
            return 0
        return 1

    return 0



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""iperf commands"""

from nest.engine.mptcp import mptcp_preload_string
from .exec import exec_exp_commands


def run_iperf_server(ns_name, server_options, is_mptcp, out, err):
    """
    Run Iperf Server on a namesapce

    Parameters
    ----------
    ns_name : str
        name of the server namespace
    server_options : str
        Specific options (like port_no, interval ) to run iperf3 command with
    is_mptcp : bool
        boolean to determine if connection is MPTCP enabled
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    mptcpize_prefix = mptcp_preload_string(is_mptcp)

    # Runs server
    return_code = exec_exp_commands(
        f"{mptcpize_prefix}ip netns exec {ns_name} iperf3 -s {server_options}",
        stdout=out,
        stderr=err,
    )
    # return code 1 denotes that server is terminated at the end of experiment by cleaning process
    # so it is not an error.
    if return_code == 1:
        return_code = 0
    return return_code


# pylint: disable=too-many-arguments
def run_iperf_client(ns_name, iperf3_options, destination_ip, ipv6, is_mptcp, out, err):
    """
    Run Iperf Client

    Parameters
    ----------
    ns_name : str
        name of the client namespace
    iperf3_options : str
        Specific options (like run_time, number_of_flows) to run iperf3 command with
    destination_ip : str
        the ip of server to which it has to connect
    ipv6 : bool
        determines if destination_ip is ipv4/ipv6
    is_mptcp : bool
        boolean to determine if connection is MPTCP enabled
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    mptcpize_prefix = mptcp_preload_string(is_mptcp)

    if ipv6:
        return exec_exp_commands(
            f"{mptcpize_prefix}"
            f"ip netns exec {ns_name} iperf3 -6 -c {destination_ip} {iperf3_options}",
            stdout=out,
            stderr=err,
        )

    return exec_exp_commands(
        f"{mptcpize_prefix}"
        f"ip netns exec {ns_name} iperf3 -c {destination_ip} {iperf3_options}",
        stdout=out,
        stderr=err,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" Util functions """

import importlib

from .exec import exec_subprocess


def is_dependency_installed(tool):
    """
    util to check if a tool is installed

    Parameters
    ----------
    tool : str
        tool to check

    Returns
    -------
    bool
        true if the `tool` is installed

    """
    return_code = exec_subprocess(f"type {tool}", shell=True)
    return return_code == 0


def is_package_installed(package):
    """
    Utility function to check if a python package is installed.
    This is done using importlib.

    Parameters
    ----------
    package : str
        package name to check for

    Returns
    -------
    bool
        true if the `package` is installed
    """
    is_pkg_present = False

    # Use importlib to find if the package exists
    spec = importlib.util.find_spec(package)

    # If package exists, return True
    if spec is not None:
        is_pkg_present = True

    return is_pkg_present



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""MPTCP commands"""

import logging
import re
import sys
from .exec import exec_subprocess, exec_subprocess_in_background

logger = logging.getLogger(__name__)


def check_compatibility(func):
    """
    Decorator to check if the system supports MPTCP
    """

    def wrapper(*args, **kwargs):
        if not compatible_iproute2_version():
            logger.error("This system does not support MPTCP.")
            sys.exit()
        return func(*args, **kwargs)

    return wrapper


def mptcp_preload_string(is_mptcp):
    """
    MPTCP Preload String

    Parameters
    ----------
    is_mptcp: bool
        boolean to determine if connection is MPTCP enabled

    Returns
    -------
    str:
        The mptcpize run preload string
    """
    return "mptcpize run " if is_mptcp else ""


@check_compatibility
def set_mptcp_node_parameters(max_subflows, max_add_addr_accepted, ns_name):
    """
    Set MPTCP node parameters

    Parameters
    ----------
    max_subflows: int
    max_add_addr_accepted: int
    ns_name: str
    """
    exec_subprocess(
        f"ip -n {ns_name} mptcp limits set "
        f"subflow {max_subflows} add_addr_accepted {max_add_addr_accepted} ",
        output=True,
    )


@check_compatibility
def add_mptcp_monitor(ns_name, folder):
    """
    Add and log MPTCP monitor
    """
    exec_subprocess_in_background(
        f"ip netns exec {ns_name} ip mptcp monitor "
        f"> {re.escape(folder)}/mptcp_monitor_{ns_name}.log &",
        shell=True,
    )


@check_compatibility
def enable_mptcp_node(ns_name):
    """
    Enable MPTCP Node

    Parameters
    ----------
    ns_name: str
    """
    exec_subprocess(
        f"ip netns exec {ns_name} sysctl -w net.mptcp.enabled=1",
        output=True,
    )
    exec_subprocess(
        f"ip -n {ns_name} mptcp endpoint flush",
        output=True,
    )


@check_compatibility
def disable_mptcp_node(ns_name):
    """
    Disable MPTCP Node

    Parameters
    ----------
    ns_name: str
    """
    exec_subprocess(
        f"ip netns exec {ns_name} sysctl -w net.mptcp.enabled=0",
        output=True,
    )
    exec_subprocess(
        f"ip -n {ns_name} mptcp endpoint flush",
        output=True,
    )


@check_compatibility
def enable_mptcp_endpoint(ns_name, interface, flags):
    """
    Set interface as an MPTCP endpoint

    Parameters
    ----------
    ns_name: str
    dev_name: Interface
    flags[]: An array of Strings
    """
    if len(set(flags) - set(possible_mptcp_flags)) == 0:
        exec_subprocess(
            f"ip -n {ns_name} "
            f"mptcp endpoint add {interface.get_address().get_addr(with_subnet=False)} "
            f"dev {interface.id} {' '.join(flags)}",
            output=True,
        )
    else:
        logger.error("Not a valid parameter")


@check_compatibility
def get_mptcp_endpoints(ns_name, interface=None):
    """
    Gets the MPTCP endpoints of the interface

    Parameters
    ----------
    ns_name: str
    interface: Interface
    """
    raw_output = exec_subprocess(
        f"ip -n {ns_name} mptcp endpoint show",
        output=True,
    )

    endpoint_matches = dict(
        list(
            map(
                lambda endpoint: [
                    re.match(r"(.*) id", endpoint).group(1),
                    re.findall(r"(subflow|signal|backup|fullmesh)", endpoint),
                ],
                re.findall(
                    r".*dev " + (interface.id if interface else r"\w+"), raw_output
                ),
            )
        )
    )

    return endpoint_matches


def __get_iproute2_version():
    """
    Get iproute2 version

    Returns
    -------
    str:
        The extracted version from the terminal output
    """
    output_string = exec_subprocess("ip -V", output=True)
    version_match = re.search(r"iproute2-(\d*.\d*.\d*)", output_string)
    if version_match:
        return version_match.group(1)
    return None


def compatible_iproute2_version():
    """
    Compare the installed iproute2 version with the threshold.
    Versions above 5.15.0 support the ip-mptcp API.

    Returns
    -------
    bool:
        False if the installed version is older than 5.15.0
    """
    threshold_version = (5, 15, 0)
    version_on_system = __get_iproute2_version()

    try:
        parsed_version_tuple = tuple(map(int, version_on_system.split(".")))
        if parsed_version_tuple < threshold_version:
            return False

    except ValueError:
        return False

    return True


def __old_iproute2_version():
    """
    Compare the installed iproute2 version with the threshold.
    Versions above 5.19.0 support fullmesh as an endpoint flag for MPTCP.

    Returns
    -------
    bool:
        True if the installed version is older than 5.19.0
    """
    threshold_version = (5, 19, 0)
    version_on_system = __get_iproute2_version()
    if version_on_system is None:
        return True
    return tuple(map(int, version_on_system.split("."))) < threshold_version


@check_compatibility
def get_default_mptcp_flags():
    """
    Set default mptcp flags depending on the version.
    Default value set to "subflow", "signal" if version < 5.19.0
    else default set to "fullmesh", "signal"

    Returns
    -------
    Array:
         consisting of default mptcp flags
    """
    return (
        ["signal", "subflow"]
        if __old_iproute2_version()
        else ["subflow", "fullmesh", "signal"]
    )


def __get_all_possible_flags():
    """
    Get all mptcp flags depending on the version.
    Returns
    -------
    Array:
         returns an array of version compatible flags
    """
    if not compatible_iproute2_version():
        return []
    if __old_iproute2_version():
        return ["subflow", "signal", "backup"]
    return ["subflow", "signal", "backup", "fullmesh"]


possible_mptcp_flags = __get_all_possible_flags()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""SIP commands"""

import subprocess
import signal
from .exec import exec_subprocess, exec_exp_commands


def run_sip_server(ns_name, port, scenario="basic", xml_file=None):
    """
    Run SIP server on provided namespace

    Parameters
    ----------
    ns_name : str
        Name of the server namespace
    port: int
        Port number to run the server on.
    scenario : str
        Specifies the scenario to be run for the experiment.
        (uses "basic" as default)
    xml_file : Path
        Specifies the path to an XML scenario file for the SIPp test.
        (Used only when scenario is "xml")
    Returns
    -------
    int
        return code of the command executed
    """
    if scenario == "branch":
        scenario_option = "-sn branchs"
    elif scenario == "xml":
        scenario_option = f"-sf {xml_file}"
    else:
        scenario_option = "-sn uas"

    cmd_string = f"""ip netns exec {ns_name} sipp -p {port}\
    {scenario_option} -bg"""

    return exec_subprocess(cmd_string)


# pylint: disable=too-many-arguments
def run_sip_client(
    ns_name,
    destination_ip,
    port,
    duration,
    out,
    err,
    scenario="basic",
    xml_file=None,
    callrate=None,
):
    """
    Run SIP client

    Parameters
    ----------
    ns_name : str
        Name of the client namespace
    destination_ip : str
        IP address of the server namespace
    port : int
        port number of the server at which the SIP application is running
    duration : int
        Specifies the duration of the SIPp test.
    out : Path
        Path to the file where standard output will be redirected.
    err : Path
        Path to the file where standard error will be redirected.
    scenario : string
        Specifies the scenario to be run for the experiment
        (uses "basic" as default)
    xml_file : Path
        Specifies the path to an XML scenario file for the SIPp test.
        (Used only when scenario is "xml")
    Returns
    -------
    int
        return code of the command executed
    """

    if scenario == "branch":
        scenario_option = "-sn branchc"
    elif scenario == "xml":
        scenario_option = f"-sf {xml_file}"
    else:
        scenario_option = "-sn uac"

    callrate_option = ""
    if callrate is not None:
        callrate_option = "-r " + str(callrate)

    cmd_string = f"""ip netns exec {ns_name} \
    sipp {scenario_option} {destination_ip}:{port} -trace_stat \
    -stf {out.name} {callrate_option}
    """

    return exec_exp_commands(
        cmd_string,
        stdout=subprocess.PIPE,
        stderr=err,
        timeout=duration,
        signal_on_timeout=signal.SIGTERM,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""Test APIs from topology sub-package"""

import os
from pathlib import Path
import unittest
import sys
import time
import json
from nest.topology import Node, Router, connect
from nest.topology.network import Network
from nest.topology.address_helper import AddressHelper
from nest.experiment import (
    Experiment,
    Flow,
    CoapApplication,
    MpegDashApplication,
    SipApplication,
)
from nest.clean_up import delete_namespaces, delete_encoded_mpeg_dash_chunks
from nest.topology_map import TopologyMap
from nest.mpeg_dash_encoder import MpegDashEncoder
from nest import config

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
class TestExperiment(unittest.TestCase):
    def test_experiment(self, qdisc=None):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", qdisc)
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        if qdisc:
            exp.require_qdisc_stats(r_n1)

        exp.run()

    def test_experiment_udp(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-udp")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_udp_flow(flow)

        exp.run()

    def test_experiment_tcp_module_params(self):
        config.set_value("show_tcp_module_parameter_confirmation", False)

        n1 = Node("n1")
        n2 = Node("n2")
        r = Node("r")
        r.enable_ip_forwarding()

        (n1_r, r_n1) = connect(n1, r)
        (r_n2, n2_r) = connect(r, n2)

        n1_r.set_address("10.1.1.1/24")
        r_n1.set_address("10.1.1.2/24")
        r_n2.set_address("10.1.2.2/24")
        n2_r.set_address("10.1.2.1/24")

        n1.add_route("DEFAULT", n1_r)
        n2.add_route("DEFAULT", n2_r)

        n1_r.set_attributes("100mbit", "5ms")
        r_n1.set_attributes("100mbit", "5ms")

        r_n2.set_attributes("1mbit", "10ms", "pfifo")
        n2_r.set_attributes("1mbit", "10ms")

        flow = Flow(n1, n2, n2_r.get_address(), 0, 20, 1)

        exp1 = Experiment("cubic-default-params")
        exp1.add_tcp_flow(flow)
        exp1.run()

        exp2 = Experiment("cubic-beta=1000")
        exp2.configure_tcp_module_params("cubic", beta=1000)
        exp2.add_tcp_flow(flow)
        exp2.run()

        config.set_value("show_tcp_module_parameter_confirmation", True)

    def test_experiment_tcp_iperf3(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-tcp-iperf3")
        flow = Flow(n0, n1, n1_r.address, 0, 10, 3)
        exp.add_tcp_flow(flow, "cubic", "iperf3")

        exp.run()

    # Test `CoapApplication` API by generating GET and PUT CoAP traffic
    def test_experiment_coap(self):
        h1 = Node("h1")
        h2 = Node("h2")
        r = Router("r")

        n1 = Network("192.168.1.0/24")
        n2 = Network("192.168.3.0/24")

        (eth1, etr1) = connect(h1, r, network=n1)
        (etr2, eth2) = connect(r, h2, network=n2)

        AddressHelper.assign_addresses()

        eth1.set_attributes("1000mbit", "1ms")
        etr1.set_attributes("1000mbit", "1ms")
        etr2.set_attributes("1000mbit", "1ms")
        eth2.set_attributes("1000mbit", "1ms")

        h1.add_route("DEFAULT", eth1)
        h2.add_route("DEFAULT", eth2)

        exp = Experiment("test-experiment-coap")

        # Number of CON and NON messages to send
        n_con_msgs = 10
        n_non_msgs = 10

        # Configure a flow from `h1` to `h2`.
        application_get = CoapApplication(
            h1, h2, eth2.get_address(), n_con_msgs, n_non_msgs
        )
        application_put = CoapApplication(
            h1, h2, eth2.get_address(), n_con_msgs, n_non_msgs
        )

        # Add the above flows as CoAP flows to the current experiment
        exp.add_coap_application(application_get)
        exp.add_coap_application(application_put)

        # Run the experiment
        exp.run()

    # Test `MpegDashApplication` API by streaming MPEG-DASH video
    # pylint: disable=too-many-locals,logging-not-lazy
    def test_experiment_mpeg_dash_vlc(self):
        current_dir = Path(os.path.abspath(__file__)).parent
        video_path = current_dir / "video.mp4"
        output_path = current_dir / "test-mpeg-dash-encoded-chunks-vlc"

        mpeg_dash_encoder = MpegDashEncoder()
        encoder_response = mpeg_dash_encoder.encode_video(
            video_path, output_path, overwrite=False
        )
        if encoder_response != 0:
            sys.exit(0)

        config.set_value("mpeg_dash_delete_encoded_chunks_on_termination", True)

        h1 = Node("h1")
        h2 = Node("h2")
        r = Router("r")

        n1 = Network("192.168.1.0/24")
        n2 = Network("192.168.2.0/24")

        (eth1, etr1) = connect(h1, r, network=n1)
        (etr2, eth2) = connect(r, h2, network=n2)

        AddressHelper.assign_addresses()

        eth1.set_attributes("10mbit", "10ms")
        etr1.set_attributes("10mbit", "10ms")
        etr2.set_attributes("5mbit", "5ms")
        eth2.set_attributes("5mbit", "5ms")

        eth1.set_packet_corruption("2%", "0.5%")
        eth2.set_packet_corruption("2%", "0.5%")

        eth1.set_packet_loss("2%")
        eth2.set_packet_loss("2%")

        h1.add_route("DEFAULT", eth1)
        h2.add_route("DEFAULT", eth2)

        exp = Experiment("test-experiment-mpeg-dash-vlc")

        # Configure a video stream flow from `h1` to `h2`
        mpegDashApplication = MpegDashApplication(
            h1,
            h2,
            eth1.get_address(),
            eth2.get_address(),
            8000,
            output_path,
            40,
            "vlc",
            enable_audio_playback=False,
        )

        # Add the above application as an MPEG-DASH flow to the current experiment
        exp.add_mpeg_dash_application(mpegDashApplication)

        # Run the experiment
        exp.run()

    def test_experiment_mpeg_dash_gpac(self):
        current_dir = Path(os.path.abspath(__file__)).parent
        video_path = current_dir / "video.mp4"
        output_path = current_dir / "test-mpeg-dash-encoded-chunks-gpac"

        mpeg_dash_encoder = MpegDashEncoder()
        encoder_response = mpeg_dash_encoder.encode_video(
            video_path, output_path, overwrite=False
        )
        if encoder_response != 0:
            sys.exit(0)

        config.set_value("mpeg_dash_delete_encoded_chunks_on_termination", True)

        h1 = Node("h1")
        h2 = Node("h2")
        r = Router("r")

        n1 = Network("192.168.1.0/24")
        n2 = Network("192.168.2.0/24")

        (eth1, etr1) = connect(h1, r, network=n1)
        (etr2, eth2) = connect(r, h2, network=n2)

        AddressHelper.assign_addresses()

        eth1.set_attributes("10mbit", "10ms")
        etr1.set_attributes("10mbit", "10ms")
        etr2.set_attributes("5mbit", "5ms")
        eth2.set_attributes("5mbit", "5ms")

        eth1.set_packet_corruption("2%", "0.5%")
        eth2.set_packet_corruption("2%", "0.5%")

        eth1.set_packet_loss("2%")
        eth2.set_packet_loss("2%")

        h1.add_route("DEFAULT", eth1)
        h2.add_route("DEFAULT", eth2)

        exp = Experiment("test-experiment-mpeg-dash-gpac")

        # Configure a video stream flow from `h1` to `h2`
        mpegDashApplication = MpegDashApplication(
            h1,
            h2,
            eth1.get_address(),
            eth2.get_address(),
            8000,
            output_path,
            40,
            "gpac",
            enable_audio_playback=False,
        )

        # Add the above application as an MPEG-DASH flow to the current experiment
        exp.add_mpeg_dash_application(mpegDashApplication)

        # Run the experiment
        exp.run()

    def test_experiment_sip(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Router("r")

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms")
        n1_r.set_attributes("10mbit", "40ms")

        exp_name = "test_experiment_sip"
        exp = Experiment(exp_name)
        duration = 60
        # Configure a SIP Application from `n0` to `n1`
        sipApplication = SipApplication(
            n0,
            n1,
            n0_r.get_address(),
            n1_r.get_address(),
            5050,
            duration,
            "basic",
        )

        # Add the above application as a SIP flow to the current experiment
        exp.add_sip_application(sipApplication)
        exp_start_time = time.localtime()
        # Run the experiment
        exp.run()

        current_dir = os.getcwd()
        timestamp = time.strftime("%d-%m-%Y-%H:%M:%S", exp_start_time)
        dump_folder = f"{exp_name}({timestamp})_dump"
        with open(os.path.join(current_dir, dump_folder, "sip.json")) as dump_file:
            stats = json.load(dump_file)
            self.assertNotEqual(stats["n0"][0]["SuccessfulCall(C)"], "0")

            dump_file.close()

    def test_experiment_sip_packet_loss(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Router("r")

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms")
        n1_r.set_attributes("10mbit", "40ms")

        exp_name = "test_experiment_sip_packet_loss"
        exp = Experiment(exp_name)
        duration = 60
        # Configure a SIP Application from `n0` to `n1`
        sipApplication = SipApplication(
            n0,
            n1,
            n0_r.get_address(),
            n1_r.get_address(),
            5050,
            duration,
            "basic",
        )

        # Add the above application as a SIP flow to the current experiment
        exp.add_sip_application(sipApplication)
        # Run the experiment again with 50% packet loss
        n0_r.set_packet_loss("50%")
        exp_start_time = time.localtime()

        exp.run()

        current_dir = os.getcwd()
        timestamp = time.strftime("%d-%m-%Y-%H:%M:%S", exp_start_time)
        dump_folder = f"{exp_name}({timestamp})_dump"
        with open(os.path.join(current_dir, dump_folder, "sip.json")) as dump_file:
            stats = json.load(dump_file)
            self.assertNotEqual(stats["n0"][0]["FailedCall(C)"], "0")

            dump_file.close()

    def test_experiment_ipv6(self):
        # Test IPv6 with Duplicate Address Detection (DAD) disabled
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10::1:1/122")
        r_n0.set_address("10::1:2/122")
        r_n1.set_address("10::2:2/122")
        n1_r.set_address("10::2:1/122")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-ipv6")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        exp.run()

    def test_experiment_ipv6_dad(self):
        # Test IPv6 with Duplicate Address Detection (DAD) enabled
        config.set_value("disable_dad", False)

        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10::1:1/122")
        r_n0.set_address("10::1:2/122")
        r_n1.set_address("10::2:2/122")
        n1_r.set_address("10::2:1/122")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-ipv6-dad")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        exp.run()

        # Resetting disable_dad in config
        config.set_value("disable_dad", True)

    def test_experiment_qdisc_codel(self):
        self.test_experiment("codel")

    def test_experiment_qdisc_fq_codel(self):
        self.test_experiment("fq_codel")

    def test_experiment_qdisc_pfifo(self):
        self.test_experiment("pfifo")

    def test_experiment_qdisc_pie(self):
        self.test_experiment("pie")

    def test_experiment_qdisc_fq_pie(self):
        self.test_experiment("fq_pie")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()
        delete_encoded_mpeg_dash_chunks()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Class to handle static routing related functionalities"""

from nest.topology_map import TopologyMap


class StaticRouting:
    """
    Handles Static Routing related fuctionalities
    """

    def __init__(self):
        # reverse mapping from interface to node, used during DFS
        self._interface_to_node_map = {}

        for host in TopologyMap.get_hosts():
            for interface in host.interfaces:
                self._interface_to_node_map[interface.address] = host

        for router in TopologyMap.get_routers():
            for interface in router.interfaces:
                self._interface_to_node_map[interface.address] = router

    def dfs(self, cur_node, dest_address, visited):
        """
        Run dfs for a given node and populate routing table entries keeping
        this node as destination. dfs also works if the topology has cycles,
        the algorithm will use a spanning tree for routing.

        Parameters
        ----------
        cur_node : Node
            The current node at which dfs is to be run
        start_node : Node
            The root node from which the dfs is run initially
        visited : Set(Node._id)
            The set of ids of all nodes visited during the dfs
        """

        for interface in cur_node.interfaces:
            interface_pair = interface.pair
            next_node = self._interface_to_node_map[interface_pair.address]

            if next_node.id not in visited:
                visited.add(next_node.id)
                next_node.add_route(
                    dest_addr=dest_address,
                    via_interface=interface_pair,
                    next_hop_addr=interface.address,
                )
                self.dfs(next_node, dest_address, visited)

    def run_static_routing(self):
        """
        Iterate through all hosts and call dfs for them
        """

        for start_node in TopologyMap.get_hosts():
            if len(start_node.interfaces) == 0:
                # a single host not connected to any router or host
                continue

            # we assume that a single host will be connected to only one router
            self.dfs(
                cur_node=start_node,
                dest_address=start_node.interfaces[0].address,
                visited=set(),
            )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Base class for all interfaces"""

import logging
import multiprocessing
import sys
import time
from nest.input_validator import input_validator
from nest.input_validator.metric import Bandwidth, Delay, Distribution, Percentage
from nest.topology.device import Ifb, Device
from nest.engine.mptcp import get_default_mptcp_flags, possible_mptcp_flags

logger = logging.getLogger(__name__)

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=invalid-name


class BaseInterface:
    """
    Abstraction for an network interface.

    Attributes
    ----------
    name : str
        User given name for the interface
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node : Node
        `Node` which contains this `Interface`
    address : str/Address
        IP address assigned to this interface
    """

    @input_validator
    def __init__(self, interface_name: str, device: Device):
        """
        Constructor of BaseInterface.

        Parameters
        ----------
        interface_name : str
            Name of the interface
        """

        self._name = interface_name
        self._device = device
        self._ifb = None

        # Track bandwidth and delay
        self._bandwidth = None
        self._delay = None

    @property
    def name(self):
        """Getter for name"""
        return self._device.name

    @property
    def id(self):
        """Getter for id"""
        return self._device.id

    @property
    def node_id(self):
        """
        Getter for the `Node` associated
        with the interface
        """
        return self._device.node_id

    @node_id.setter
    def node_id(self, node_id):
        """
        Setter for the `Node` associated
        with the interface

        Parameters
        ----------
        node : Node
            The node where the interface is to be installed
        """
        self._device.node_id = node_id

    @property
    def ifb_id(self):
        """
        Getter for the id of the ifb of
        the interface
        """
        return self._ifb.id

    @property
    def address(self):
        """
        NOTE: This method is deprecated and will be removed in future versions.
              Use get_address method instead.

        Getter for the address associated
        with the interface
        """
        return self._device.get_address()

    @address.setter
    def address(self, address):
        """
        NOTE: This method is deprecated and will be removed in future versions.
              Use set_address method instead.

        Assigns IP address to an interface

        Parameters
        ----------
        address : Address/str or List[Address/str]
            IP address to be assigned to the interface
        """
        self._device.set_address(address)

    def set_address(self, address):
        """
        Assigns IP addresses to an interface

        Parameters
        ----------
        address : Address/str or List[Address/str]
            IP address to be assigns to the interface
        """
        self._device.set_address(address)

    def add_address(self, address):
        """
        Adds IP addresses to an interface

        Parameters
        ----------
        address : Address/str or List[Address/str]
            IP address to be added to the interface
        """
        self._device.add_address(address)

    def del_address(self, address):
        """
        Delete IP address(es) from the interface

        Parameters
        ----------
        address : str or list
            IP address to be deleted from the interface
        """
        self._device.del_address(address)

    def enable_mpls(self):
        """
        Enables mpls input through the interface.
        Requires mpls kernel modules to be loaded.

        Run ``sudo modprobe mpls_iptunnel`` to load mpls modules.
        """
        self._device.enable_mpls()

    def is_mpls_enabled(self):
        """
        Check if the interface has mpls enabled
        """
        return self._device.is_mpls_enabled()

    # pylint: disable=dangerous-default-value
    def enable_mptcp_endpoint(self, flags=get_default_mptcp_flags()):
        """
        Declares the interface as an MPTCP endpoint
        """
        if len(set(flags) - set(possible_mptcp_flags)) > 0:
            logger.error(
                "The mptcp flags are missing or invalid for your iproute2 version"
            )
            sys.exit()
        self._device.enable_mptcp_endpoint(flags)

    # pylint: disable=dangerous-default-value
    def get_mptcp_endpoints(self):
        """
        Gets the MPTCP endpoints of the interface
        """
        return self._device.get_mptcp_endpoints()

    @property
    def mtu(self):
        """
        Get the maximum transmit unit value for the interface
        """
        return self._device.mtu

    @mtu.setter
    def mtu(self, mtu_value):
        """
        Set the maximum transmit unit value for the interface
        Default is 1500 bytes.
        """
        self._device.mtu = mtu_value

    @input_validator
    def get_address(self, ipv4: bool = True, ipv6: bool = True, as_list: bool = False):
        """
        Gets the required IP addresses for the interface
        Returns a list or an Address object

        Parameters
        ----------
        ipv4 : If set to true, the IPv4 address of the interface is returned (defaults to True)
        ipv6 : If set to true, the IPv6 address of the interface is returned (defaults to True)
        If both are True, both the addresses are returned
        Either ipv4 or ipv6 must be True
        as_list : Only applicable when a single address is set (Applicable individually for
                  ipv4 and ipv6).
        Returns Address object when false, else returns a single Address object in list
        (defaults to False).
        """
        return self._device.get_address(ipv4, ipv6, as_list)

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for an interface.
        """
        self._device.disable_ip_dad()

    @input_validator
    def set_mode(self, mode: str, delay: float = 0.0):
        """
        Changes the mode of the interface

        Parameters
        ----------
        mode : string
            interface mode to be set
        delay : float
            time in sec after which the mode is set
        """
        if delay >= 0:
            time.sleep(delay)
            self._device.set_mode(mode)
        else:
            raise ValueError(f"{delay} should be greater than or equal to 0")

    def get_qdisc(self):
        """
        Note that this is the qdisc set inside
        the IFB.
        """
        if self._ifb is not None:
            for qdisc in self._ifb.qdisc_list:
                if qdisc.parent == "1:1" and qdisc.handle == "11:":
                    return qdisc
        return None

    def _create_and_mirred_to_ifb(self):
        """
        Creates a IFB for the interface so that a Qdisc can be
        installed on it
        Mirrors packets to be sent out of the interface first to
        itself (IFB)
        Assumes the interface has already invoked _set_structure()

        Parameters
        ----------
        """

        ifb_name = "ifb-" + self.name
        self._ifb = Ifb(ifb_name, self.node_id, self.id)

    @input_validator
    def set_bandwidth(self, bandwidth: Bandwidth):
        """
        Sets a minimum bandwidth for the interface
        It is done by adding a HTB qdisc and a rate parameter to the class

        Parameters
        ----------
        bandwidth : Bandwidth
            The minimum rate that has to be set in kbit
        """

        self._device.set_structure()
        self._bandwidth = bandwidth.string_value

        bandwidth_parameter = {"rate": bandwidth.string_value}

        self._device.change_class("htb", "1:", "1:1", **bandwidth_parameter)

    @input_validator
    def set_delay(self, delay: Delay):
        """
        Adds a delay to the link between two namespaces
        It is done by adding a delay in the interface

        Parameters
        ----------
        delay : Delay
            The delay to be added
        """
        # TODO: Make adding delay possible without bandwidth being set

        self._device.set_structure()
        self._delay = delay

        delay_parameter = {"delay": delay.string_value}

        self._device.change_qdisc("11:", "netem", **delay_parameter)

    @input_validator
    def set_delay_distribution(
        self, delay: Delay, jitter: Delay, distribution: Distribution
    ):
        """
        Allows to choose delay distribution. If not specified,
        the default delay distribution is normal.

        Parameters:
        ------------
        delay: Delay
            Add delay to the outgoing packets on a specified interface
        jitter: Jitter
            Variations of delay
        distribution: Distribution
            Default delay distribution: normal | pareto | paretonormal | experimental
        """

        self._delay = delay + self._delay

        distribution_parameter = {
            "delay": self._delay.string_value,
            "": jitter.string_value,
            "distribution": distribution.option,
        }

        self._device.set_structure()

        self._device.change_qdisc("11:", "netem", **distribution_parameter)

    @input_validator
    def set_packet_corruption(
        self, corrupt_rate: Percentage, correlation_rate: Percentage = None
    ):
        """
        allows the emulation of random noise introducing an error in a
        random position for a chosen percent of packets.
        It is also possible to add a correlation.

        Parameters
        ----------
        corrupt_rate : str
            rate of the packets to be corrupted
        correlation_rate : str
            correlation between the corrupted packets
        """
        parsed_correlation_rate = (
            correlation_rate.string_value if correlation_rate else ""
        )

        self._device.set_structure()

        corrupt_parameter = {
            "corrupt": corrupt_rate.string_value,
            "": parsed_correlation_rate,
        }

        self._device.change_qdisc("11:", "netem", **corrupt_parameter)

    @input_validator
    def set_packet_loss(
        self, loss_rate: Percentage, correlation_rate: Percentage = None
    ):
        """
        adds an independent loss probability to the packets outgoing from
        the chosen network interface.
        It is also possible to add a correlation

        Parameters
        ----------
        loss_rate : str
            rate of the packets to be lost
        correlation_rate : str
            correlation between the lost packets
        """
        parsed_correlation_rate = (
            correlation_rate.string_value if correlation_rate else ""
        )

        self._device.set_structure()

        loss_parameter = {"loss": loss_rate.string_value, "": parsed_correlation_rate}

        self._device.change_qdisc("11:", "netem", **loss_parameter)

    @input_validator
    def set_packet_loss_state(
        self,
        p13: Percentage,
        p31: Percentage = None,
        p32: Percentage = None,
        p23: Percentage = None,
        p14: Percentage = None,
        ecn: bool = False,
    ):
        """
        Adds packet loss using Markov model with 4-states
        of transitional probabilities as parameters
        p13 p31 p23 p32 p14 in sequence. p13 is mandatory.

        Parameters
        -----------
        p13: Percentage
            State 1, corresponds to good reception.
        p31: Percentage
            State 2, from Burst to Good state, shows second state.
        p23: Percentage
            Third state to burst loss of packet
        p32: Percentage
            Third state, that corresponds to burst loss of packet
        p14: Percentage
            Fourth state, corresponds to independent loss of packet
        ecn: bool
            To marked packet instead of dropping them
        """

        probabilities = f"{p13.string_value} "

        if p31 is not None:
            probabilities += f"{p31.string_value} "
        else:
            probabilities = "0% "

        if p32 is not None:
            probabilities += f"{p32.string_value} "
        else:
            probabilities += "0% "

        if p23 is not None:
            probabilities += f"{p23.string_value} "
        else:
            probabilities += "0% "

        if p14 is not None:
            probabilities += f"{p14.string_value} "
        else:
            probabilities += "0% "

        if ecn is True:
            probabilities += "ecn"

        loss_parameter = {"loss": "state", "": probabilities}

        self._device.set_structure()

        self._device.change_qdisc("11:", "netem", **loss_parameter)

    @input_validator
    def set_packet_loss_gemodel(
        self,
        p: Percentage,
        r: Percentage = None,
        h: Percentage = None,
        k: Percentage = None,
        ecn: bool = False,
    ):
        """
        adds packet loss using Gilber Elliot Model,
        To use Bernoulli model, only parameter p is required.
        Simple Gilbert Model requires p and r parameters.
        Gilbert model needs p, r and 1-h parameter
        Gilber Elliot model require all four, i.e p, r , 1-h and 1-k

        Parameter
        -----------
        p: Percentage
            Probability from Good state to Burst State
        r: Percentage
            Probability from Burst state to Good State
        h: Percentage
            Probability of packet transmision on Burst State
        k: Percentage
            Probability of pracket loss in Good state
        """

        loss_parameters = {"loss": "gemodel"}

        probabilities = f"{p.string_value} "

        if r is not None:
            probabilities += f"{r.string_value} "
        else:
            probabilities += "0% "

        if h is not None:
            probabilities += f"{h.string_value} "
        else:
            probabilities += "0% "

        if k is not None:
            probabilities += f"{k.string_value} "
        else:
            probabilities += "0% "

        if ecn is True:
            probabilities += "ecn"

        loss_parameters[""] = probabilities

        self._device.set_structure()

        self._device.change_qdisc("11:", "netem", **loss_parameters)

    @input_validator
    def set_packet_duplication(self, duplicate_rate: Percentage):
        """
        Using this option the chosen percent of packets is duplicated
        before queuing them.

        Paramters
        ----------
        duplicate_rate: Percentage
            rate of packets to get duplicated.

        """

        self._device.set_structure()

        duplicate_parameter = {"duplicate": duplicate_rate.string_value}

        self._device.change_qdisc("11:", "netem", **duplicate_parameter)

    @input_validator
    def set_packet_reordering(
        self, delay: Delay, reorder_rate: Percentage, gap: int = None
    ):
        """
        Adds reordering of packets configured on qdisc to outgoing packets
        from one host to another.

        Parameter
        ---------
        delay: Delay
            Delay to be added for packet reordering
        reorder_rate: Percentage
            Percentage of packets to reorder
        gap : int (Optional)
            (gap - 1) packets will be delayed, and subsequent packets
            will be sent immediately.
        """

        new_delay = delay + self._delay

        reorder_parameter = {
            "delay": new_delay.string_value,
            "reorder": reorder_rate.string_value,
        }

        if gap is not None:
            reorder_parameter["gap"] = str(gap)

        self._device.set_structure()

        self._device.change_qdisc("11:", "netem", **reorder_parameter)

    def set_qdisc(self, qdisc, **kwargs):
        """
        Adds the Queueing algorithm to the interface

        Parameters
        ----------
        qdisc : string
            The Queueing algorithm to be added
        bandwidth :
            Link bandwidth
        """

        qdisc_list = ["choke", "codel", "fq_codel", "pie", "fq_pie", "pfifo", "red"]

        if qdisc not in qdisc_list:
            raise ValueError(f"{qdisc} is not a valid Qdisc")

        self._device.set_structure()

        if self._ifb is None:
            self._create_and_mirred_to_ifb()

        if self._bandwidth is not None:
            self._ifb.set_bandwidth(self._bandwidth)

        self._ifb.delete_qdisc("11:")
        self._ifb.add_qdisc(qdisc, "1:1", "11:", **kwargs)

    @input_validator
    def set_attributes(
        self, bandwidth: Bandwidth, delay: Delay, qdisc: str | None = None, **kwargs
    ):
        """
        Add attributes bandwidth, delay and qdisc to interface

        Parameters
        ----------
        bandwidth : str/Bandwidth
            Packet outgoing rate
        delay : str/Delay
            Delay before packet is sent out
        qdisc : string
            The Queueing algorithm to be added to interface
            (Default value = None)
        """

        self.set_bandwidth(bandwidth)
        self.set_delay(delay)

        if qdisc is not None:
            self.set_qdisc(qdisc, **kwargs)

    def enable_offload(self, offload_name):
        """
        API for enabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to enable
        """

        self._device.enable_offload(offload_name)

    def disable_offload(self, offload_name):
        """
        API for disabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to disable
        """

        self._device.disable_offload(offload_name)

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"

    @input_validator
    def disable(self, start_time: float = -1.0, end_time: float = -1.0):
        """
        API for disable a network interface from 'start_time' sec  to 'end_time' sec

        Parameters
        ----------
        start_time : float
            time in sec after which the interface mode is set 'DOWN'
        end_time : float
            time in sec after which the interface mode is set 'UP'
        """
        if start_time == -1.0 and end_time == -1.0:
            self.set_mode("DOWN")

        elif start_time >= 0.0 and end_time == -1.0:
            disable_process = multiprocessing.Process(
                target=self.set_mode, args=["DOWN", start_time]
            )
            disable_process.start()

        elif start_time == -1.0 and end_time > 0.0:
            raise ValueError(
                "start_time is required and should be greater than or equal to 0"
            )

        elif start_time >= 0.0 and end_time > 0.0:
            if start_time < end_time:
                disable_process = multiprocessing.Process(
                    target=self.set_mode, args=["DOWN", start_time]
                )
                enable_process = multiprocessing.Process(
                    target=self.set_mode, args=["UP", end_time]
                )
                disable_process.start()
                enable_process.start()
            else:
                raise ValueError(f"{start_time} should be smaller than {end_time}")
        else:
            raise ValueError(
                "start_time should be greater than or equal to 0 "
                "and end_time should be greater than 0"
            )

    @input_validator
    def enable(self, start_time: float = -1.0, end_time: float = -1.0):
        """
        API for enable a network interface from 'start_time' sec to 'end_time' sec

        Parameters
        ----------
        start_time : float
            time in second after which the interface mode is set 'UP'
        end_time : float
            time in second which the interface mode is set 'DOWN'
        """
        if start_time == -1.0 and end_time == -1.0:
            self.set_mode("UP")

        elif start_time >= 0.0 and end_time == -1.0:
            enable_process = multiprocessing.Process(
                target=self.set_mode, args=["UP", start_time]
            )
            enable_process.start()

        elif start_time == -1.0 and end_time > 0.0:
            raise ValueError(
                "start_time is required and should be greater than or equal to 0"
            )

        elif start_time >= 0.0 and end_time > 0.0:
            if start_time < end_time:
                enable_process = multiprocessing.Process(
                    target=self.set_mode, args=["UP", start_time]
                )
                disable_process = multiprocessing.Process(
                    target=self.set_mode, args=["DOWN", end_time]
                )
                enable_process.start()
                disable_process.start()
            else:
                raise ValueError(f"{start_time} should be smaller than {end_time}")
        else:
            raise ValueError(
                "start_time should be greater than or equal to 0 "
                "and end_time should be greater than 0"
            )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""Script to be run for running experiments on topology"""

from multiprocessing import Process
from collections import namedtuple, defaultdict
import logging
import os
import sys
from time import sleep
import copy
from tqdm import tqdm
from nest.engine.mptcp import add_mptcp_monitor

from nest.logging_helper import DepedencyCheckFilter
from nest import config
from nest.topology_map import TopologyMap
from nest.clean_up import kill_processes, tcp_modules_clean_up
from nest import engine
from .pack import Pack

# Import results
from .results import (
    Iperf3ServerResults,
    SsResults,
    NetperfResults,
    Iperf3Results,
    TcResults,
    PingResults,
    CoAPResults,
    MpegDashResults,
    SipResults,
)

# Import parsers
from .parser.ss import SsRunner
from .parser.netperf import NetperfRunner
from .parser.iperf3 import Iperf3Runner, Iperf3ServerRunner
from .parser.tc import TcRunner
from .parser.ping import PingRunner
from .parser.coap import CoAPRunner
from .parser.mpeg_dash import MpegDashRunner

# Import plotters
from .plotter.ss import plot_ss
from .plotter.netperf import plot_netperf
from .plotter.iperf3 import plot_iperf3
from .plotter.tc import plot_tc
from .plotter.ping import plot_ping
from .plotter.mpeg_dash import plot_mpeg_dash
from ..engine.util import is_dependency_installed, is_package_installed
from .parser.sip import SipRunner

logger = logging.getLogger(__name__)
if not any(isinstance(filter, DepedencyCheckFilter) for filter in logger.filters):
    # Duplicate filter is added to avoid logging of same error
    # messages incase any of the tools is not installed
    logger.addFilter(DepedencyCheckFilter())


# TODO:Refactor code so as to not exceed 1000 lines
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals, too-many-branches
# pylint: disable=too-many-statements, invalid-name
def run_experiment(exp):
    """
    Run experiment

    Parameters
    -----------
    exp : Experiment
        The experiment attributes
    """

    tcp_modules_helper(exp)
    additional_tools = ["mptcpize"]
    tools = [
        "netperf",
        "ss",
        "tc",
        "iperf3",
        "ping",
        "coap",
        "server",
        "mpeg_dash",
        "sip",
    ]
    Runners = namedtuple("runners", tools)
    exp_runners = Runners(
        netperf=[],
        ss=[],
        tc=[],
        iperf3=[],
        ping=[],
        coap=[],
        server=[],
        mpeg_dash=[],
        sip=[],
    )  # Runner objects

    # Keep track of all destination nodes [to ensure netperf, iperf3 and
    # coap server is run at most once]
    destination_nodes = {
        "netperf": set(),
        "iperf3": set(),
        "coap": set(),
        "mpeg_dash": set(),
        "sip": set(),
    }

    # Contains start time and end time to run respective command
    # from a source netns to destination address (in destination netns)
    ss_schedules = defaultdict(lambda: (float("inf"), float("-inf")))
    ping_schedules = defaultdict(lambda: (float("inf"), float("-inf")))

    # Overall experiment stop time considering all flows
    exp_end_t = float("-inf")

    dependencies = get_dependency_status(exp, tools + additional_tools)

    ss_required = False
    ss_filters = set()
    server_runner = []

    # Traffic generation
    for flow in exp.flows:
        iperf3_options = {}
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            dst_addr,
            start_t,
            stop_t,
            _,
            options,
            _,
        ] = flow._get_props()  # pylint: disable=protected-access

        exp_end_t = max(exp_end_t, stop_t)

        (min_start, max_stop) = ping_schedules[(src_ns, dst_ns, dst_addr)]
        ping_schedules[(src_ns, dst_ns, dst_addr)] = (
            min(min_start, start_t),
            max(max_stop, stop_t),
        )

        # Setup TCP/UDP flows
        if options["protocol"] in ["TCP", "MPTCP"]:
            ss_required = True

            if options["tool"] == "netperf":
                # * Ignore netperf tcp control connections
                # * Destination port of netperf control connection is 12865
                # * We also have "sport" (source port) in the below condition since
                #   there can be another flow in the reverse direction whose control
                #   connection also we must ignore.
                ss_filters.add("sport != 12865 and dport != 12865")

                (tcp_runners, ss_schedules,) = setup_tcp_flows(
                    {
                        k: v
                        for k, v in dependencies.items()
                        if k in ["netperf", "mptcpize"]
                    },
                    flow,
                    ss_schedules,
                    destination_nodes["netperf"],
                    options["protocol"] == "MPTCP",
                )
                exp_runners.netperf.extend(tcp_runners)
                # Update destination nodes
                destination_nodes["netperf"].add(dst_ns)

            elif options["tool"] == "iperf3":
                ss_filters.add("sport != 5201 and dport != 5201")
                (tcp_runners, ss_schedules,) = setup_tcp_flows(
                    {
                        k: v
                        for k, v in dependencies.items()
                        if k in ["iperf3", "mptcpize"]
                    },
                    flow,
                    ss_schedules,
                    destination_nodes["iperf3"],
                    options["protocol"] == "MPTCP",
                )
                exp_runners.iperf3.extend(tcp_runners)
                # Update destination nodes
                destination_nodes["iperf3"].add(dst_ns)

                dst_port_options = {}
                port_nos = options["port_nos"]
                for port_no in port_nos:
                    options["port_no"] = port_no
                    dst_port_options[port_no] = copy.deepcopy(options)

                if dst_ns in iperf3_options:
                    dst_port_options.update(iperf3_options.get(dst_ns))
                iperf3_options.update({dst_ns: dst_port_options})

        elif options["protocol"] == "udp":
            # * Ignore iperf3 tcp control connections
            # * Destination port of iperf3  control connection is 5201
            # * We also have "sport" (source port) in the below condition since
            #   there can be another flow in the reverse direction whose control
            #   connection also we must ignore.
            ss_filters.add("sport != 5201 and dport != 5201")
            udp_runners = setup_udp_flows(dependencies["iperf3"], flow)

            exp_runners.iperf3.extend(udp_runners)

            # Update destination nodes
            destination_nodes["iperf3"].add(dst_ns)
            dst_port_options = {options["port_no"]: options}
            if dst_ns in iperf3_options:
                dst_port_options.update(iperf3_options.get(dst_ns))
            iperf3_options.update({dst_ns: dst_port_options})

        server_runner.extend(
            run_server(
                iperf3_options,
                exp_end_t,
                options["protocol"],
                options["protocol"] == "MPTCP",
            )
        )

    for coap_application in exp.coap_applications:
        [
            src_ns,
            dst_ns,
            dst_addr,
            _,
            _,
            _,
        ] = coap_application._get_props()  # pylint: disable=protected-access

        config.set_value("show_progress_bar", False)

        # Setup runners for emulating CoAP traffic
        coap_runners = setup_coap_runners(
            dependencies["coap"], coap_application, destination_nodes["coap"]
        )
        exp_runners.coap.extend(coap_runners)
        destination_nodes["coap"].add(dst_ns)

    for mpeg_dash_application in exp.mpeg_dash_applications:
        [
            _,
            dst_ns,
            _,
            dst_addr,
            port,
            _,
            duration,
            _,
            _,
            _,
        ] = mpeg_dash_application._get_props()  # pylint: disable=protected-access

        exp_end_t = max(exp_end_t, duration)

        # Setup runners for emulating Mpeg-Dash traffic
        (mpeg_dash_runners, ss_schedules) = setup_mpeg_dash_runners(
            dependencies["mpeg_dash"],
            mpeg_dash_application,
            ss_schedules,
            destination_nodes["mpeg_dash"],
        )
        exp_runners.mpeg_dash.extend(mpeg_dash_runners)
        destination_nodes["mpeg_dash"].add((dst_ns, dst_addr, port))

        if len(exp.mpeg_dash_applications) > 0:
            if ss_filters == set():
                ss_required = True
                # Excluding some states from ss statistics which
                # do not contain any internal TCP information.
                ss_filters.add("exclude fin-wait-2 exclude time-wait exclude syn-recv")

    for sip_application in exp.sip_applications:
        [
            src_ns,
            dst_ns,
            _,
            dst_addr,
            port,
            duration,
            _,
            _,
            _,
            _,
        ] = sip_application._get_props()  # pylint: disable=protected-access

        exp_end_t = max(exp_end_t, duration)

        # Setup runners for emulating SIP traffic
        sip_runners = setup_sip_runners(
            dependencies["sip"],
            sip_application,
            destination_nodes["sip"],
        )
        exp_runners.sip.extend(sip_runners)
        destination_nodes["sip"].add((dst_ns, dst_addr, port))

    if ss_required:
        ss_filter = " and ".join(ss_filters)
        ss_runners = setup_ss_runners(dependencies["ss"], ss_schedules, ss_filter)
        exp_runners.ss.extend(ss_runners)

    tc_runners = setup_tc_runners(dependencies["tc"], exp.qdisc_stats, exp_end_t)
    exp_runners.tc.extend(tc_runners)

    ping_runners = setup_ping_runners(dependencies["ping"], ping_schedules)
    exp_runners.ping.extend(ping_runners)

    try:
        # Start traffic generation
        run_workers(setup_flow_workers(exp_runners, exp_end_t))

        logger.info("Parsing statistics...")

        exp_runners.server.extend(server_runner)

        # Parse the stored statistics
        run_workers(setup_parser_workers(exp_runners))

        logger.info("Parsing statistics complete!")
        logger.info("Output results as JSON dump...")

        # Output results as JSON dumps
        dump_json_outputs()

        if config.get_value("readme_in_stats_folder"):
            # Copying README.txt to stats folder
            relative_path = os.path.join("info", "README.txt")
            readme_path = os.path.join(os.path.dirname(__file__), relative_path)
            Pack.copy_files(readme_path)

        if config.get_value("plot_results"):
            logger.info("Plotting results...")

            # Plot results and dump them as images
            run_workers(setup_plotter_workers())

            logger.info("Plotting complete!")

        logger.info("Experiment %s complete!", exp.name)

    except KeyboardInterrupt:
        logger.warning(
            "Experiment %s forcefully stopped. The results obtained maybe incomplete!",
            exp.name,
        )
    finally:
        cleanup()


def tcp_modules_helper(exp):
    """
    This function is called at the beginning of run_experiment
    to perform tcp modules related helper tasks

    Parameters
    -----------
    exp : Experiment
        The experiment attributes
    """
    if exp.tcp_module_params:
        if (
            not (config.get_value("show_tcp_module_parameter_confirmation"))
            or input(
                "Are you sure you want to modify TCP module parameters in Linux kernel? (y/n) : "
            ).lower()
            == "y"
        ):
            for cong_algo, params in exp.tcp_module_params.items():
                flag = engine.is_module_loaded(cong_algo)
                if flag:
                    # the module is already loaded, so store the old parameters
                    # during experiment set these parameters with new values (reset=False)
                    # during cleanup reset these parameters with old values (reset=True)
                    exp.old_cong_algos[cong_algo] = engine.get_current_params(cong_algo)
                    engine.set_tcp_params(cong_algo, params, False)
                else:
                    # the module will be newly loaded
                    # it should be removed during cleanup
                    (exp.new_cong_algos).append(cong_algo)
                    params_string = " ".join(
                        {f"{key}={value}" for key, value in params.items()}
                    )
                    engine.load_tcp_module(cong_algo, params_string)


def run_server(iperf3options, exp_end_t, protocol, is_mptcp):
    """
    Run and wait for all server to start

    Parameters
    ----------
    iperf3options: dict
        start server with iperf3 server options
    exp_end_t: int
        experiment completion time
    protocol: str
        transport layer protocol, either "tcp" or "udp"
    """
    # Start server
    server_list = []
    for dst_ns in iperf3options:
        for dst_port in iperf3options[dst_ns]:
            runner_obj = Iperf3ServerRunner(dst_ns, exp_end_t, protocol, is_mptcp)
            runner_obj.setup_iperf3_server(iperf3options[dst_ns][dst_port])
            server_list.append(runner_obj)

    for server in server_list:
        process = Process(target=server.run)
        process.start()

    return server_list


def run_workers(workers):
    """
    Run and wait for processes to finish

    Parameters
    ----------
    workers: list[multiprocessing.Process]
        List of processes to be run
    """
    # Start workers
    for worker in workers:
        worker.start()

    # wait for all the workers to finish
    for worker in workers:
        worker.join()


def setup_plotter_workers():
    """
    Setup plotting processes

    Returns
    -------
    List[multiprocessing.Process]
        plotters
    """
    plotters = []

    plotters.append(Process(target=plot_ss, args=(SsResults.get_results(),)))
    plotters.append(Process(target=plot_netperf, args=(NetperfResults.get_results(),)))
    plotters.append(Process(target=plot_iperf3, args=(Iperf3Results.get_results(),)))
    plotters.append(Process(target=plot_tc, args=(TcResults.get_results(),)))
    plotters.append(Process(target=plot_ping, args=(PingResults.get_results(),)))
    plotters.append(
        Process(target=plot_mpeg_dash, args=(MpegDashResults.get_results(),))
    )

    return plotters


def dump_json_outputs():
    """
    Outputs experiment results as json dumps
    """
    SsResults.output_to_file()
    NetperfResults.output_to_file()
    Iperf3Results.output_to_file()
    TcResults.output_to_file()
    PingResults.output_to_file()
    CoAPResults.output_to_file()
    Iperf3ServerResults.output_to_file()
    MpegDashResults.output_to_file()
    SipResults.output_to_file()


def setup_flow_workers(exp_runners, exp_stop_time):
    """
    Setup flow generation and stats collection processes(netperf, ss, tc, iperf3...).

    Also add a progress bar process for showing experiment progress.

    Parameters
    ----------
    exp_runners: collections.NamedTuple
        all(netperf, ping, ss, tc..) the runners
    exp_stop_time: int
        Time when experiment stops (in seconds)

    Returns
    -------
    List[multiprocessing.Process]
        flow generation and stats collection processes
        + progress bar process
    """
    workers = []

    for runners in exp_runners:
        workers.extend([Process(target=runner.run) for runner in runners])

    # Add progress bar process
    if config.get_value("show_progress_bar"):
        workers.extend([Process(target=progress_bar, args=(exp_stop_time,))])

    return workers


def setup_parser_workers(exp_runners):
    """
    Setup parsing processes

    Parameters
    ----------
    exp_runners: collections.NamedTuple
        all(netperf, ping, ss, tc..) the runners

    Returns
    -------
    List[multiprocessing.Process]
        parsers
    """
    parsers = []

    for ss_runner in exp_runners.ss:
        parsers.append(Process(target=ss_runner.parse))

    for netperf_runner in exp_runners.netperf:
        parsers.append(Process(target=netperf_runner.parse))

    for iperf3_runner in exp_runners.iperf3:
        parsers.append(Process(target=iperf3_runner.parse))

    for tc_runner in exp_runners.tc:
        parsers.append(Process(target=tc_runner.parse))

    for ping_runner in exp_runners.ping:
        parsers.append(Process(target=ping_runner.parse))

    for coap_runner in exp_runners.coap:
        parsers.append(Process(target=coap_runner.parse))

    for mpeg_dash_runner in exp_runners.mpeg_dash:
        parsers.append(Process(target=mpeg_dash_runner.parse))

    for server_runner in exp_runners.server:
        parsers.append(Process(target=server_runner.parse))

    for sip_runner in exp_runners.sip:
        parsers.append(Process(target=sip_runner.parse))

    return parsers


def get_dependency_status(exp, tools):
    """
    Checks for dependency

    Parameters
    ----------
    exp : Experiment
        The experiment attributes
    tools: List[str]
        list of tools to check for it's installation

    Returns
    -------
    dict
        contains information as to whether `tools` are installed
    """
    dependencies = {}
    for dependency in tools:
        # Check for the availability of aiocoap for CoAP emulation
        if dependency == "coap":
            dependencies[dependency] = is_package_installed("aiocoap")
            continue
        # Check for the availability of the media players for MPEG-DASH emulation
        if dependency == "mpeg_dash":
            media_players_specified_in_exp = {
                app.player for app in exp.mpeg_dash_applications
            }
            missing_media_players = [
                media_player
                for media_player in media_players_specified_in_exp
                if not is_dependency_installed(media_player)
            ]
            if missing_media_players:
                raise RuntimeError(
                    "The following media player(s) are specified in ",
                    f"the experiment but aren't installed: {', '.join(missing_media_players)}",
                )
            dependencies[dependency] = True
            continue
        # Check for the availability of sipp for SIP emulation
        if dependency == "sip":
            dependencies[dependency] = is_dependency_installed("sipp")
            continue
        dependencies[dependency] = is_dependency_installed(dependency)
    return dependencies


def setup_tcp_flows(
    dependencies, flow, ss_schedules, destination_nodes, is_mptcp=False
):
    """
    Setup netperf/iperf3 to run tcp flows
    Parameters
    ----------
    dependencies: dictionary of dependencies
        whether dependencies are installed
    flow: Flow
        Flow parameters
    ss_schedules:
        ss_schedules so far
    destination_nodes:
        Destination nodes so far already running netperf/iperf3 server
    is_mptcp:
        boolean to determine if connection is MPTCP enabled

    Returns
    -------
    dependency: int
        updated dependency in case netperf/iperf3 is not installed
    tcp_runners: List[NetperfRunner] / List[Iperf3Runner]
        all the netperf/iperf3 flows generated
    workers: List[multiprocessing.Process]
        Processes to run netperf/iperf3 flows
    ss_schedules: dict
        updated ss_schedules
    """
    tcp_runners = []

    # pylint: disable=use-a-generator
    if not all(
        [
            status if dependency in ["netperf", "iperf3"] else True
            for dependency, status in dependencies.items()
        ]
    ):
        logger.error("Netperf/iperf3 not found. Tcp flows cannot be generated")
        sys.exit()
    if is_mptcp and not dependencies["mptcpize"]:
        logger.error("Mptcpize not found. MPTCP connections cannot be made")
        sys.exit()
    # Get flow attributes
    [
        src_ns,
        dst_ns,
        dst_addr,
        start_t,
        stop_t,
        n_flows,
        options,
        src_addr,
    ] = flow._get_props()  # pylint: disable=protected-access

    source_node = TopologyMap.get_node(src_ns)
    destination_node = TopologyMap.get_node(dst_ns)

    if source_node.mptcp_monitor_required:
        add_mptcp_monitor(src_ns, Pack.FOLDER)

    if destination_node.mptcp_monitor_required:
        add_mptcp_monitor(dst_ns, Pack.FOLDER)

    if options["tool"] == "netperf":
        # Run netserver if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            NetperfRunner.run_netserver(dst_ns, is_mptcp)

        src_name = TopologyMap.get_node(src_ns).name
        netperf_options = {}
        netperf_options["testname"] = "TCP_STREAM"
        netperf_options["cong_algo"] = options["cong_algo"]
        f_flow = "flow" if n_flows == 1 else "flows"
        logger.info(
            "Running %s netperf %s from %s (%s) to %s...",
            n_flows,
            f_flow,
            src_name,
            src_addr if src_addr else "Default IP",
            dst_addr,
        )

        # Create new processes to be run simultaneously
        for _ in range(n_flows):
            runner_obj = NetperfRunner(
                src_ns,
                src_addr,
                dst_addr,
                start_t,
                stop_t - start_t,
                dst_ns,
                is_mptcp,
                **netperf_options,
            )
            tcp_runners.append(runner_obj)

    elif options["tool"] == "iperf3":
        src_name = TopologyMap.get_node(src_ns).name
        f_flow = "flow" if n_flows == 1 else "flows"
        iperf3_options = {}
        iperf3_options["cong_algo"] = options["cong_algo"]

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-function-args
        logger.info(
            "Running %s tcp iperf3 %s from %s (%s) to %s...",
            n_flows,
            f_flow,
            src_name,
            src_addr if src_addr else "Default IP",
            dst_addr,
        )

        port_nos = options["port_nos"]

        # Create new processes to be run simultaneously
        for i in range(n_flows):
            runner_obj = Iperf3Runner(
                src_ns,
                dst_addr,
                options.get("target_bw"),
                1,
                start_t,
                stop_t - start_t,
                dst_ns,
                "tcp",
                is_mptcp,
                **iperf3_options,
            )
            options["port_no"] = port_nos[i]
            runner_obj.setup_iperf3_client(options)
            tcp_runners.append(runner_obj)

    # Since MPTCP potentially has multiple destination addresses,
    # we will run SS on each of those to ensure that user has visibility
    # into each destination interface's statistics.
    if options["protocol"] in ["MPTCP"]:
        for interface in flow.destination_node.interfaces:
            ss_schedules = _get_start_stop_time_for_ss(
                src_ns,
                dst_ns,
                interface.get_address().get_addr(with_subnet=False),
                start_t,
                stop_t,
                ss_schedules,
            )
    else:
        ss_schedules = _get_start_stop_time_for_ss(
            src_ns, dst_ns, dst_addr, start_t, stop_t, ss_schedules
        )
    return tcp_runners, ss_schedules


def setup_udp_flows(dependency, flow):
    """
    Setup iperf3 to run udp flows

    Parameters
    ----------
    dependency: int
        whether iperf3 is installed
    flow: Flow
        Flow parameters
    destination_nodes:
        Destination nodes so far already running iperf3 server

    Returns
    -------
    dependency: int
        updated dependency in case iproute2 is not installed
    iperf3_runners: List[NetperfRunner]
        all the iperf3 udp flows generated
    workers: List[multiprocessing.Process]
        Processes to run iperf3 udp flows
    """
    iperf3_runners = []
    if not dependency:
        logger.error("Iperf3 not found. Udp flows cannot be generated")
        sys.exit()
    # Get flow attributes
    [
        src_ns,
        dst_ns,
        dst_addr,
        start_t,
        stop_t,
        n_flows,
        options,
        src_addr,
    ] = flow._get_props()  # pylint: disable=protected-access

    src_name = TopologyMap.get_node(src_ns).name
    f_flow = "flow" if n_flows == 1 else "flows"
    logger.info(
        "Running %s udp %s from %s (%s) to %s...",
        n_flows,
        f_flow,
        src_name,
        src_addr if src_addr else "Default IP",
        dst_addr,
    )

    runner_obj = Iperf3Runner(
        src_ns,
        dst_addr,
        options["target_bw"],
        n_flows,
        start_t,
        stop_t - start_t,
        dst_ns,
        "udp",
    )
    runner_obj.setup_iperf3_client(options)
    iperf3_runners.append(runner_obj)

    return iperf3_runners


def setup_ss_runners(dependency, ss_schedules, ss_filter):
    """
    setup SsRunners for collecting tcp socket statistics

    Parameters
    ----------
    dependency: int
        whether ss is installed
    ss_schedules: dict
        start time and end time for SsRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[SsRunners]
    """
    runners = []
    if dependency:
        logger.info("Running ss on nodes...")
        for key, timings in ss_schedules.items():
            src_ns = key[0]
            dst_ns = key[1]
            dst_addr = key[2]
            ss_runner = SsRunner(
                src_ns,
                dst_addr,
                timings[0],
                timings[1] - timings[0],
                dst_ns,
                ss_filter=ss_filter,
            )
            runners.append(ss_runner)
    else:
        logger.warning("ss not found. Sockets stats will not be collected")
    return runners


def setup_tc_runners(dependency, qdisc_stats, exp_end):
    """
    setup TcRunners for collecting qdisc statistics

    Parameters
    ----------
    dependency: int
        whether tc is installed
    qdisc_stats: dict
        info regarding nodes to run tc on
    exp_end: float
        time to stop running tc
    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run tc at nodes
    runners: List[TcRunners]
    """
    runners = []
    if dependency and len(qdisc_stats) > 0:
        logger.info("Running tc on requested interfaces...")
        for qdisc_stat in qdisc_stats:
            tc_runner = TcRunner(
                qdisc_stat["ns_id"], qdisc_stat["int_id"], qdisc_stat["qdisc"], exp_end
            )
            runners.append(tc_runner)
    elif not dependency:
        logger.warning("tc not found. Qdisc stats will not be collected")
    return runners


def setup_ping_runners(dependency, ping_schedules):
    """
    setup PingRunners for collecting latency

    Parameters
    ----------
    dependency: int
        whether ping is installed
    ping_schedules: dict
        start time and end time for PingRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[PingRunner]
    """
    runners = []
    if dependency:
        for key, timings in ping_schedules.items():
            src_ns = key[0]
            dst_ns = key[1]
            dst_addr = key[2]
            ping_runner = PingRunner(
                src_ns, dst_addr, timings[0], timings[1] - timings[0], dst_ns
            )
            runners.append(ping_runner)
    else:
        logger.warning("ping not found.")
    return runners


def setup_coap_runners(dependency, application, destination_nodes):
    """
    Setup CoAPRunner objects for generating CoAP traffic

    Parameters
    ----------
    dependency : int
        Whether aiocoap is installed
    application : CoapApplication
        The CoapApplication object
    destination_nodes:
        Destination nodes so far already running CoAP server

    Returns
    -------
    runners : List[CoAPRunner]
        List of CoAPRunner objects for the current flow object
    """
    runners = []

    # If aiocoap is installed
    if dependency:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            dst_addr,
            n_con_msgs,
            n_non_msgs,
            user_options,
        ] = application._get_props()  # pylint: disable=protected-access

        # Run CoAP server if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            # If user has not supplied the user options
            if user_options is not None:
                # Creating the options string for running the CoAP server
                if (
                    "coap_server_content" in user_options.keys()
                    and user_options["coap_server_content"] != ""
                ):
                    server_content = '"' + user_options["coap_server_content"] + '"'
                    server_options = f"-c {server_content}"
                else:
                    server_options = None
            else:
                server_options = None

            # Running the server
            CoAPRunner.run_server(dst_ns, server_options)

        # Create the CoAPRunner object
        coap_runner = CoAPRunner(src_ns, dst_addr, user_options, n_con_msgs, n_non_msgs)
        runners.append(coap_runner)

    # If aiocoap is not installed
    else:
        logger.warning("aiocoap not found for CoAP emulation.")

    # Return the list of runners
    return runners


def setup_mpeg_dash_runners(dependency, application, ss_schedules, destination_nodes):
    """
    Setup MpegDashRunner objects for generating Mpeg-Dash traffic

    Parameters
    ----------
    dependency : int
        Whether dependencies required for MPEG-DASH emulation are installed or not.
    application : MpegDashApplication
        The MpegDashApplication object
    destination_nodes:
        Server nodes so far already running Mpeg-Dash server

    Returns
    -------
    (runners,ss_schedules) : Tuple
        runners : List[MpegDashRunner]
            List of MpegDashRunner objects for the current flow object
        ss_schedules: List
            List containing ss schedule information
    """
    runners = []

    # If the required media players are installed
    if dependency:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            src_addr,
            dst_addr,
            port,
            encoded_chunks_path,
            duration,
            player,
            enable_audio_playback,
            additional_player_options,
        ] = application._get_props()  # pylint: disable=protected-access

        # Run Mpeg-Dash server if not already run before on given dst_ns, dst_addr and port.
        if (dst_ns, dst_addr, port) not in destination_nodes:
            MpegDashRunner.run_server(dst_ns, port, encoded_chunks_path)

        # Create the MpegDashRunner object
        mpeg_dash_runner = MpegDashRunner(
            src_ns,
            dst_addr,
            dst_ns,
            port,
            encoded_chunks_path,
            duration,
            player,
            enable_audio_playback,
            additional_player_options,
        )
        runners.append(mpeg_dash_runner)

        ss_schedules = _get_start_stop_time_for_ss(
            dst_ns, src_ns, src_addr, 0, duration, ss_schedules
        )
    else:
        logger.warning(
            """Dependencies required for MPEG-DASH emulation aren't installed. """
        )

    # Return the list of runners and ss schedules
    return runners, ss_schedules


def setup_sip_runners(dependency, application, destination_nodes):
    """
    Setup SipRunner objects for generating SIP traffic

    Parameters
    ----------
    dependency : int
        Whether dependencies required for SIP emulation are installed or not.
    application : SipApplication
        The SipApplication object
    destination_nodes:
        Server nodes so far already running SIP server

    Returns
    -------
    runners : List[SipRunner]
        List of SipRunner objects for the current flow object
    """
    runners = []
    if dependency:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            _,
            dst_addr,
            port,
            duration,
            scenario,
            server_xml_file,
            client_xml_file,
            callrate,
        ] = application._get_props()  # pylint: disable=protected-access

        # Create the SipRunner object
        sip_runner = SipRunner(
            src_ns,
            dst_addr,
            dst_ns,
            port,
            duration,
            scenario,
            server_xml_file,
            client_xml_file,
            callrate=callrate,
        )

        # Run SIP server if not already run before on given dst_ns, dst_addr and port.
        if (dst_ns, dst_addr, port) not in destination_nodes:
            sip_runner.run_server(dst_ns, port, scenario, server_xml_file)

        runners.append(sip_runner)
    else:
        logger.warning("""Dependencies required for SIP emulation aren't installed. """)

    # Return the list of runners
    return runners


def progress_bar(stop_time, precision=1):
    """
    Show a progress bar from from 0 `units` to `stop_time`

    The time unit is decided by `precision` in seconds. It is
    1s by default.

    Parameters
    -----------
    stop_time : int
        The time needed 100% completion
    precision : int
        Time unit for updating progress bar. 1 second be default
    """
    try:
        print()
        for _ in tqdm(range(0, stop_time, precision), desc="Experiment Progress"):
            sleep(precision)
        print()
    except KeyboardInterrupt:
        logger.debug(
            "ProgressBar process received KeyboardInterrupt. Stopping it gracefully."
        )

    logger.info("Cleaning up all the spawned child processes...")


def cleanup():
    """
    Clean up
    """
    # Remove results of the experiment
    SsResults.remove_all_results()
    NetperfResults.remove_all_results()
    TcResults.remove_all_results()
    PingResults.remove_all_results()
    CoAPResults.remove_all_results()
    Iperf3Results.remove_all_results()
    Iperf3ServerResults.remove_all_results()
    MpegDashResults.remove_all_results()
    SipResults.remove_all_results()

    # Clean up the configured TCP modules and kill processes
    tcp_modules_clean_up()
    kill_processes()


# Helper methods
# pylint: disable=too-many-arguments
def _get_start_stop_time_for_ss(
    src_ns, dst_ns, dst_addr, start_t, stop_t, ss_schedules
):
    """
    Find the start time and stop time to run ss command in node `src_ns`
    to a `dst_addr`

    Parameters
    ----------
    src_ns: str
        ss run from `src_ns`
    dst_ns: str
        destination network namespace for ss
    dst_addr: str
        Destination address
    start_t: int
        Start time of ss command
    stop_t: int
        Stop time of ss command
    ss_schedules: list
        List with ss command schedules

    Returns
    -------
    List: Updated ss_schedules
    """
    if (src_ns, dst_ns, dst_addr) not in ss_schedules:
        ss_schedules[(src_ns, dst_ns, dst_addr)] = (start_t, stop_t)
    else:
        (min_start, max_stop) = ss_schedules[(src_ns, dst_ns, dst_addr)]
        ss_schedules[(src_ns, dst_ns, dst_addr)] = (
            min(min_start, start_t),
            max(max_stop, stop_t),
        )

    return ss_schedules
